# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import projection
from nemos.third_party.jaxopt.jaxopt import ProjectedGradient
from nemos.third_party.jaxopt.jaxopt import ScipyBoundedMinimize
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


N_CALLS = 0

class ProjectedGradientTest(test_util.JaxoptTestCase):

  def test_non_negative_least_squares(self):
    rng = onp.random.RandomState(0)
    X = rng.randn(10, 5)
    w = rng.rand(5)
    y = jnp.dot(X, w)
    fun = objective.least_squares
    w_init = jnp.zeros_like(w)

    pg = ProjectedGradient(fun=fun,
                           projection=projection.projection_non_negative)
    pg_sol = pg.run(w_init, data=(X, y)).params

    lbfgsb = ScipyBoundedMinimize(fun=fun, method="l-bfgs-b")
    lower_bounds = jnp.zeros_like(w_init)
    upper_bounds = jnp.ones_like(w_init) * jnp.inf
    bounds = (lower_bounds, upper_bounds)
    lbfgsb_sol = lbfgsb.run(w_init, bounds=bounds, data=(X, y)).params

    self.assertArraysAllClose(pg_sol, lbfgsb_sol, atol=1e-2)

  def test_projected_gradient_l2_ball(self):
    rng = onp.random.RandomState(0)
    X = rng.randn(10, 5)
    w = rng.rand(5)
    y = jnp.dot(X, w)
    fun = objective.least_squares
    w_init = jnp.zeros_like(w)

    pg = ProjectedGradient(fun=fun,
                           projection=projection.projection_l2_ball)
    pg_sol = pg.run(w_init, hyperparams_proj=1.0, data=(X, y)).params
    self.assertLess(jnp.sqrt(jnp.sum(pg_sol ** 2)), 1.0)

  def test_projected_gradient_l2_ball_manual_loop(self):
    rng = onp.random.RandomState(0)
    X = rng.randn(10, 5)
    w = rng.rand(5)
    y = jnp.dot(X, w)
    fun = objective.least_squares
    params = jnp.zeros_like(w)

    pg = ProjectedGradient(fun=fun,
                           projection=projection.projection_l2_ball)

    state = pg.init_state(params)

    for _ in range(10):
      params, state = pg.update(params, state, hyperparams_proj=1.0, data=(X, y))

    self.assertLess(jnp.sqrt(jnp.sum(params ** 2)), 1.0)

  def test_projected_gradient_implicit_diff(self):
    rng = onp.random.RandomState(0)
    X = rng.randn(10, 5)
    w = rng.rand(5)
    y = jnp.dot(X, w)
    fun = objective.least_squares
    w_init = jnp.zeros_like(w)

    def solution(radius):
      pg = ProjectedGradient(fun=fun,
                             projection=projection.projection_l2_ball)
      return pg.run(w_init, hyperparams_proj=radius, data=(X, y)).params

    eps = 1e-4
    J = jax.jacobian(solution)(0.1)
    J2 = (solution(0.1 + eps) - solution(0.1 - eps)) / (2 * eps)
    self.assertArraysAllClose(J, J2, atol=1e-2)

  def test_polyhedron_projection(self):
    def f(x):
      return x[0]**2-x[1]**2

    A = jnp.array([[0, 0]])
    b = jnp.array([0])
    G = jnp.array([[-1, -1], [0, 1], [1, -1], [-1, 0], [0, -1]])
    h = jnp.array([-1, 1, 1, 0, 0])
    hyperparams = (A, b, G, h)

    proj = projection.projection_polyhedron
    pg = ProjectedGradient(fun=f, projection=proj, jit=False)
    sol, state = pg.run(init_params=jnp.array([0.,1.]), hyperparams_proj=hyperparams)
    self.assertLess(state.error, pg.tol)

  @parameterized.product(n_iter=[10])
  def test_n_calls(self, n_iter):
    """Test whether the number of function calls
    is equal to the number of iterations + 1 in the
    no linesearch case, where the complexity is linear."""
    def fun(x):
      global N_CALLS
      N_CALLS += 1
      return x[0]**2-x[1]**2

    A = jnp.array([[0, 0]])
    b = jnp.array([0])
    G = jnp.array([[-1, -1], [0, 1], [1, -1], [-1, 0], [0, -1]])
    h = jnp.array([-1, 1, 1, 0, 0])
    hyperparams = (A, b, G, h)

    proj = projection.projection_polyhedron
    pg = ProjectedGradient(fun=fun, projection=proj, jit=False, maxiter=n_iter, tol=1e-10, stepsize=1.0)
    sol, state = pg.run(init_params=jnp.array([0.,1.]), hyperparams_proj=hyperparams)
    self.assertEqual(N_CALLS, n_iter)


if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quadratic programming with equality constraints only."""

from typing import Any
from typing import Callable
from typing import Optional

from functools import partial
from dataclasses import dataclass

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src import implicit_diff as idf
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_add, tree_sub, tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_vdot, tree_negative, tree_l2_norm
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_zeros_like
from nemos.third_party.jaxopt.jaxopt._src.linear_operator import _make_linear_operator
from nemos.third_party.jaxopt.jaxopt._src.cvxpy_wrapper import _check_params
import nemos.third_party.jaxopt.jaxopt._src.linear_solve as linear_solve
from nemos.third_party.jaxopt.jaxopt._src.iterative_refinement import IterativeRefinement


def extract_Qc_from_obj(init_params, params_obj, fun):
  """Returns (params_Q, c) parameters from params_obj.

  Args:
    init_params: KKTSolution used for the inference of the shape of primal variables.
      Mandatory when `fun` is not `None`.
    params_obj: tuple of parameters for the objective function.
    fun: objective function.

  When `fun` is `None` it retrieves the relevant informations from the tuple `params_obj`,
  whereas when `fun` is not `None` it extracts it using AutoDiff.
  """
  if fun is None:
    params_Q, c = params_obj
    return params_Q, c
  if init_params is None or init_params.primal is None:
    raise ValueError("init_params must be provided when fun is not None.")
  zeros = tree_zeros_like(init_params.primal)
  # f(x) := 0.5 * x^T Q x + c^T x + cste
  value_and_grad_fun = jax.value_and_grad(fun, argnums=0)
  # nabla_x f(x) = Q x + c
  # f(0)         = cste
  # Q is never computed explicitly, only its action on vectors is needed
  cste, c = value_and_grad_fun(zeros, params_obj)
  return (params_obj, c, cste), c


def _make_eq_qp_optimality_fun(matvec_Q, matvec_A, fun):
  """Makes the optimality function for quadratic programming.

  Args:
    matvec_Q: a function that computes the matrix-vector product with Q.
    matvec_A: a function that computes the matrix-vector product with A.
    fun: a function that computes the objective value.

  Returns:
    optimality_fun(params, params_obj, params_eq, params_ineq) where
      params = (primal_var, eq_dual_var, None)
      params_obj = (params_Q, c)
      params_eq = (params_A, b)
  """

  if fun is None:
    def obj_fun(primal_var, params_obj):
      params_Q, c = extract_Qc_from_obj(primal_var, params_obj, fun)
      Q = matvec_Q(params_Q)
      xQx = tree_vdot(primal_var, Q(primal_var))
      cx = tree_vdot(primal_var, c)
      return tree_add_scalar_mul(cx, 0.5, xQx)
  else:
    obj_fun = fun

  def eq_fun(primal_var, params_eq):
    params_A, b = params_eq
    A = matvec_A(params_A)
    return tree_sub(A(primal_var), b)

  optimality_fun_with_ineq = idf.make_kkt_optimality_fun(obj_fun, eq_fun, ineq_fun=None)

  # It is required to post_process the output of `idf.make_kkt_optimality_fun`
  # to make the signatures of optimality_fun() and run() agree.
  def optimality_fun(params, params_obj, params_eq):
    return optimality_fun_with_ineq(params, params_obj, params_eq, None)

  return optimality_fun


@dataclass(eq=False)
class EqualityConstrainedQP(base.Solver):
  """Quadratic programming with equality constraints only.

  Supports implicit differentiation, matvec, pytrees, and quadratic functions.
  Can benefit from GPU/TPU acceleration.

  If the algorithm diverges on some instances, it might be useful to tweak the
  ``refine_regularization`` parameter.

  Attributes:
    matvec_Q: a Callable matvec_Q(params_Q, u).
      By default, matvec_Q(Q, u) = dot(Q, u), where Q = params_Q.
      ``matvec_Q`` incompatible with the specification of ``fun``.
      The shape of primal variables may be inferred from params_obj = (matvec_Q, c).
    matvec_A: a Callable matvec_A(params_A, u).
      By default, matvec_A(A, u) = dot(A, u), where A = params_A.
    fun: (optional) a function with signature fun(params, params_obj) that is promised
      to be a quadratic polynomial convex with respect to params, i.e fun can be written ::
        fun(x, params_obj) = 0.5*jnp.dot(x, jnp.dot(Q, x)) + jnp.dot(c, x) + cste
      with params_obj a pytree that contains the parameters of the objective function.
      (Q, c) do not need to be explicited in params_obj by the user: c will be inferred by Jaxopt,
        and the operator x -> Qx will be computed upon request.
      ``fun`` incompatible with the specification of ``matvec_Q``.
      Note that the shape of primal cannot be inferred from params_obj anymore,
      so the user should provide it in init_params.
    solve: a Callable to solve linear systems, that accepts matvecs
      (default: linear_solve.solve_gmres).
    refine_regularization: a float (default: 0.) used to regularize the system.
      Useful for badly conditioned problems that lead to divergence.
      ``IterativeRefinement`` is used to correct the error introduced.
    refine_maxiter: maximum number of refinement steps, when
      refine_regularization is not 0.
    maxiter: maximum number of iterations.
    tol: tolerance (stoping criterion).
    implicit_diff: whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.
    jit: whether to JIT-compile the optimization loop (default: "auto").
  """

  matvec_Q: Optional[Callable] = None
  matvec_A: Optional[Callable] = None
  fun: Optional[Callable] = None
  solve: Callable = linear_solve.solve_gmres
  refine_regularization: float = 0.0
  refine_maxiter: int = 10
  maxiter: int = 1000
  tol: float = 1e-5
  implicit_diff_solve: Optional[Callable] = None
  jit: bool = True

  def _refined_solve(self, matvec, b, init, maxiter, tol):
    # Instead of solving S x = b
    # We solve     \bar{S} x = b
    #
    # S = [[  Q   A^T
    #         A    0 ]]
    #
    # \bar{S} = [[ Q + delta      A^T
    #                A         - delta ]]
    #
    # Since Q is PSD, and delta a diagonal matrix,
    # this makes \bar{S} a quasi-definite matrix.
    #
    # Quasi-Definite matrices are indefinite,
    # but they are never singular (when delta > 0).
    #
    # This guarantees that the system is well posed in every case,
    # even if some constraints of A are redundant.
    #
    # The particular form of this system is inspired by [2].
    # Quasi-Definite matrices are a known tool in the literature on
    # Interior Point methods [1].
    #
    # References:
    #
    #  [1] Vanderbei, R.J., 1995. Symmetric quasidefinite matrices.
    #      SIAM Journal on Optimization, 5(1), pp.100-113.
    #
    #  [2] Stellato, B., Banjac, G., Goulart, P., Bemporad, A. and Boyd, S., 2020.
    #      OSQP: An operator splitting solver for quadratic programs.
    #      Mathematical Programming Computation, 12(4), pp.637-672.

    def matvec_qp(_, x):
      return matvec(x)

    ridge = self.refine_regularization

    def matvec_regularized_qp(_, x):
      primal, dual_eq = x
      Stop, Sbottom = matvec(x)
      a = tree_add_scalar_mul(Stop, ridge, primal)
      b = tree_add_scalar_mul(Sbottom, -ridge, dual_eq)
      return a, b

    solver = IterativeRefinement(
      matvec_A=matvec_qp,
      matvec_A_bar=matvec_regularized_qp,
      solve=partial(self.solve, maxiter=maxiter, tol=tol),
      maxiter=self.refine_maxiter,
      tol=tol,
    )
    return solver.run(init_params=init, A=None, b=b)[0]

  def run(
    self,
    init_params: Optional[base.KKTSolution] = None,
    params_obj: Optional[Any] = None,
    params_eq: Optional[Any] = None,
  ) -> base.OptStep:
    """Solves 0.5 * x^T Q x + c^T x subject to Ax = b.

    This solver returns both the primal solution (x) and the dual solution.

    Args:
      init_params: (optional), used to infer the shape of primal variables.
        Mandatory if ``fun`` is not None, since the shape of primal variables
        cannot be inferred from ``params_obj``
      params_obj: parameters of the quadratic objective, can be:
        a tuple (Q, c) with Q a pytree of matrices,
        or a tuple (params_Q, c) if ``matvec_Q`` is provided,
        or an arbitrary pytree if ``fun`` is provided.
      params_eq: parameters of the equality constraints, can be:
        a tuple (A, b) with A a pytree of matrices,
        or a tuple (params_A, b) if matvec_A is provided.
    Returns:
      (params, state),  where params = (primal_var, dual_var_eq, None)
    """
    if self._check_params and False:
      _check_params(params_obj, params_eq)

    params_Q, c = extract_Qc_from_obj(init_params, params_obj, self.fun)
    params_A, b = params_eq

    Q = self.matvec_Q(params_Q)
    A = self.matvec_A(params_A)

    def matvec(u):
      primal_u, dual_u = u
      mv_A, rmv_A = A.matvec_and_rmatvec(primal_u, dual_u)
      return (tree_add(Q(primal_u), rmv_A), mv_A)

    target = (tree_negative(c), b)

    if init_params is not None:
      init_params = (init_params.primal, init_params.dual_eq)

    # Solves the following linear system:
    # [[Q A^T]  [primal_var = [-c
    #  [A 0  ]]  dual_var  ]    b]
    if self.refine_regularization == 0.0:
      primal, dual_eq = self.solve(
        matvec,
        target,
        init=init_params,
        tol=self.tol,
        maxiter=self.maxiter,
      )
    else:
      primal, dual_eq = self._refined_solve(
        matvec, target, init_params, tol=self.tol, maxiter=self.maxiter
      )

    return base.OptStep(params=base.KKTSolution(primal, dual_eq, None),
                        state=None)

  def l2_optimality_error(
    self,
    params: base.KKTSolution,
    params_obj: Optional[Any],
    params_eq: Optional[Any],
  ):
    """Computes the L2 norm of the KKT residuals."""
    tree = self.optimality_fun(params, params_obj, params_eq)
    return tree_l2_norm(tree)

  def __post_init__(self):
    self._check_params = self.matvec_Q is None and self.matvec_A is None

    if self.fun is not None and self.matvec_Q is not None:
      raise ValueError(f"Specification of parameter 'fun' is incompatible with 'matvec_Q' in method __init__ of {type(self)}")

    if self.fun is not None:
      def matvec_Q(params_obj, x):
        params_Q, c, _ = params_obj
        # f(x) = 0.5 * x^T Q x + c^T x + cste
        # nabla_x f(x) = Q x + c
        # Qx = nabla_x f(x) - c
        def fun_minus_cx(xx):
          return self.fun(xx, params_Q) - jnp.sum(c*xx)
        Qx = jax.grad(fun_minus_cx)(x)
        return Qx
      self.matvec_Q = matvec_Q

    self.matvec_Q = _make_linear_operator(self.matvec_Q)
    self.matvec_A = _make_linear_operator(self.matvec_A)

    self.optimality_fun = _make_eq_qp_optimality_fun(self.matvec_Q,
                                                     self.matvec_A,
                                                     self.fun)

    # Set up implicit diff.
    decorator = idf.custom_root(
      self.optimality_fun, has_aux=True, solve=self.implicit_diff_solve
    )
    # pylint: disable=g-missing-from-attributes
    self.run = decorator(self.run)

    if self.jit:
      self.run = jax.jit(self.run)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt import BacktrackingLineSearch
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_negative
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot

import numpy as onp

from sklearn import datasets

N_FUN_CALLS = 0
N_GRAD_CALLS = 0
class BacktrackingLinesearchTest(test_util.JaxoptTestCase):

  def _check_conditions_satisfied(
      self,
      condition_name,
      c1,
      c2,
      stepsize,
      initial_value,
      initial_grad,
      final_state):
    self.assertTrue(jnp.all(final_state.done))
    self.assertFalse(jnp.any(final_state.failed))

    descent_direction = tree_scalar_mul(-1, initial_grad)
    # Check sufficient decrease for all line search methods.
    sufficient_decrease = jnp.all(
        final_state.value <= initial_value +
        c1 * stepsize * tree_vdot(final_state.grad, descent_direction))
    self.assertTrue(sufficient_decrease)

    if condition_name == "goldstein":
      goldstein = jnp.all(
          final_state.value >= initial_value +
          (1 - c1) * stepsize * tree_vdot(
              initial_grad, descent_direction))
      self.assertTrue(goldstein)
    elif condition_name == "strong-wolfe":
      new_slope = tree_vdot(final_state.grad, descent_direction)
      slope = tree_vdot(initial_grad, descent_direction)
      curvature = jnp.all(jnp.abs(new_slope) <= c2 * jnp.abs(slope))
      self.assertTrue(curvature)

    elif condition_name == "wolfe":
      new_slope = tree_vdot(final_state.grad, descent_direction)
      slope = tree_vdot(initial_grad, descent_direction)
      curvature = jnp.all(new_slope >= c2 * slope)
      self.assertTrue(curvature)

  def _check_stepsize_match_params(self,
                                   value_and_grad_fun, 
                                   stepsize, 
                                   params, 
                                   descent_direction, 
                                   state, 
                                   **fun_kwargs):
    self.assertTrue(state.failed or state.done)
    step = tree_add_scalar_mul(params, stepsize, descent_direction)
    value, grad = value_and_grad_fun(step, **fun_kwargs)
    self.assertAllClose(step, state.params)
    self.assertAllClose(value, state.value)
    self.assertAllClose(grad, state.grad)


  def test_aux_value(self):
    X, y = datasets.make_classification(n_samples=10, n_features=5, n_classes=2,
                                        n_informative=3, random_state=0)
    data = (X, y)
    true_fun = objective.binary_logreg

    def augmented_fun(w, data):
      value = true_fun(w, data)
      aux = w[0:1]
      return value, aux

    rng = onp.random.RandomState(0)
    w_init = rng.randn(X.shape[1])
    ls = BacktrackingLineSearch(
        fun=augmented_fun, maxiter=20, condition="armijo", has_aux=True
    )
    _, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    _, aux_at_params = augmented_fun(state.params, data)
    self.assertArraysEqual(aux_at_params, state.aux)

  @parameterized.product(cond=[
      "armijo", "goldstein", "strong-wolfe", "wolfe"])
  def test_backtracking_linesearch(self, cond):
    X, y = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=2, n_informative=3, random_state=0
    )
    data = (X, y)
    fun = objective.binary_logreg

    rng = onp.random.RandomState(0)
    w_init = rng.randn(X.shape[1])
    initial_grad = jax.grad(fun)(w_init, data=data)
    initial_value = fun(w_init, data=data)

    # Manual loop.
    ls = BacktrackingLineSearch(fun=fun, condition=cond)
    stepsize = 1.0
    state = ls.init_state(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    stepsize, state = ls.update(stepsize=stepsize, state=state, params=w_init,
                                fun_kwargs={"data": data})
    error1 = state.error
    _, state = ls.update(
        stepsize=stepsize, state=state, params=w_init, fun_kwargs={"data": data}
    )
    error2 = state.error
    self.assertLessEqual(error2, error1)

    # Call to run.
    ls = BacktrackingLineSearch(fun=fun, maxiter=20, condition=cond)
    stepsize, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    self.assertLessEqual(state.error, 1e-5)
    self._check_conditions_satisfied(
        cond, ls.c1, ls.c2, stepsize, initial_value, initial_grad, state)
    self._check_stepsize_match_params(jax.value_and_grad(fun),
                                      stepsize,
                                      w_init,
                                      tree_negative(initial_grad),
                                      state, 
                                      data=data)

    # Call to run with value_and_grad=True.
    ls = BacktrackingLineSearch(fun=jax.value_and_grad(fun),
                                condition=cond,
                                maxiter=20,
                                value_and_grad=True)
    stepsize, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    self.assertLessEqual(state.error, 1e-5)
    self._check_conditions_satisfied(
        cond, ls.c1, ls.c2, stepsize, initial_value, initial_grad, state)
    self._check_stepsize_match_params(jax.value_and_grad(fun),
                                      stepsize,
                                      w_init,
                                      tree_negative(initial_grad),
                                      state,
                                      data=data)
    # Failed linesearch (high c1 ensures convergence condition is not met).
    ls = BacktrackingLineSearch(fun=fun, maxiter=20, condition=cond, c1=2.)
    stepsize, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )
    self.assertTrue(jnp.all(state.failed))
    self.assertFalse(jnp.any(state.done))
    self._check_stepsize_match_params(jax.value_and_grad(fun),
                                      stepsize,
                                      w_init,
                                      tree_negative(initial_grad),
                                      state,
                                      data=data)

  @parameterized.product(
      cond=["armijo", "goldstein", "strong-wolfe", "wolfe"],
      val=[onp.inf, onp.nan])
  def test_backtracking_linesearch_non_finite(self, cond, val):

    def fun(x):
      result = jnp.where(x > 4., val, (x - 2)**2)
      grad = jnp.where(x > 4., onp.nan, 2 * (x - 2.))
      return result, grad

    x_init = -0.001
    _, grad = fun(x_init)
    descent_direction = tree_negative(grad)

    # Make sure initial step triggers a non-finite value.
    ls = BacktrackingLineSearch(
        fun=fun, condition=cond, max_stepsize=2.0, value_and_grad=True)
    stepsize = 1.25

    stepsize, state = ls.run(init_stepsize=stepsize, params=x_init)
    self._check_stepsize_match_params(fun,
                                      stepsize,
                                      x_init,
                                      descent_direction,
                                      state)
    self.assertFalse(jnp.isnan(state.value) or jnp.isinf(state.value))

  @parameterized.product(cond=["armijo", "goldstein"])
  def test_eval_count(self, cond):
    global N_FUN_CALLS
    global N_GRAD_CALLS

    def fun(x):
      global N_FUN_CALLS
      N_FUN_CALLS += 1
      return 32*jnp.sum(x**2)
    
    _value_and_grad_fun = jax.value_and_grad(fun)

    def value_and_grad_fun(x):
      global N_GRAD_CALLS
      N_GRAD_CALLS += 1
      return _value_and_grad_fun(x)
    
    params = jnp.asarray(1.)
    init_stepsize = jnp.asarray(1.)
    # Reset number of calls
    N_FUN_CALLS = 0
    N_GRAD_CALLS = 0

    linesearch = BacktrackingLineSearch(fun,
                                        value_and_grad=value_and_grad_fun,
                                        condition=cond,
                                        jit=False)
    _, ls_state = linesearch.run(init_stepsize, params)
    self.assertEqual(N_FUN_CALLS, ls_state.num_fun_eval)
    self.assertEqual(N_GRAD_CALLS, ls_state.num_grad_eval)
    # One initial gradient call and one final = 2 calls
    self.assertEqual(N_GRAD_CALLS, 2)
    self.assertGreater(N_FUN_CALLS, N_GRAD_CALLS)


if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  #jax.config.update("jax_enable_x64", True)
  absltest.main()



"""Validation and construction of inverse link functions."""

from typing import Callable

import jax
import jax.numpy as jnp

from ..utils import one_over_x

LINK_NAME_TO_FUNC = {
    "nemos.utils.one_over_x": one_over_x,
    "jax.numpy.exp": jnp.exp,
    "jax._src.numpy.ufuncs.exp": jnp.exp,
    "jax.nn.softplus": jax.nn.softplus,
    "jax._src.nn.functions.softplus": jax.nn.softplus,
    "jax.scipy.special.expit": jax.scipy.special.expit,
    "jax._src.scipy.special.expit": jax.scipy.special.expit,
    "jax.lax.logistic": jax.lax.logistic,
    "jax._src.lax.lax.logistic": jax.lax.logistic,
    "jax.scipy.stats.norm.cdf": jax.scipy.stats.norm.cdf,
    "jax._src.scipy.stats.norm.cdf": jax.scipy.stats.norm.cdf,
    "softplus": jax.nn.softplus,
    "exp": jax.numpy.exp,
    "one_over_x": one_over_x,
    "logistic": jax.lax.logistic,
    "norm.cdf": jax.scipy.stats.norm.cdf,
    "expit": jax.scipy.special.expit,
}


def link_function_from_string(link_name: str):
    """
    Get a link function from a given name.

    Parameters
    ----------
    link_name:
        A string representation of the link function, e.g. "jax.numpy.exp".

    Returns
    -------
    :
        The link function corresponding to the provided name.

    Raises
    ------
    ValueError:
        If the provided string does not match any known link function.
    """
    if link_name in LINK_NAME_TO_FUNC:
        return LINK_NAME_TO_FUNC[link_name]
    else:
        raise ValueError(
            f"Unknown link function: {link_name}. "
            f"Link function must be one of {list(LINK_NAME_TO_FUNC.keys())}"
            f"if you want to use a custom link function, please provide it as a Callable."
            f"if you think this is a bug, please open an issue at 'https://github.com/flatironinstitute/nemos/issues'."
        )


def check_inverse_link_function(inverse_link_function: Callable):
    """
    Check if the provided inverse_link_function is usable.

    This function verifies if the inverse link function:

    1. Is callable
    2. Returns a jax.numpy.ndarray
    3. Is differentiable (via jax)

    Parameters
    ----------
    inverse_link_function :
        The function to be checked.

    Raises
    ------
    TypeError
        If the function is not callable.
    ValueError
        If the function does not return a jax.numpy.ndarray,
        or is not differentiable.
    """
    if inverse_link_function in LINK_NAME_TO_FUNC.values():
        return

    # check that it's callable
    if not callable(inverse_link_function):
        raise TypeError("The `inverse_link_function` function must be a Callable!")

    # check if the function returns a jax array for a 1D array
    array_out = inverse_link_function(jnp.array([1.0, 2.0, 3.0]))
    if not isinstance(array_out, jnp.ndarray):
        raise ValueError("The `inverse_link_function` must return a jax.numpy.ndarray!")

    # Optionally: Check for scalar input
    scalar_out = inverse_link_function(1.0)
    if not isinstance(scalar_out, (jnp.ndarray, float, int)):
        raise ValueError(
            "The `inverse_link_function` must handle scalar inputs correctly and return a scalar or a "
            "jax.numpy.ndarray!"
        )

    # check for autodiff
    try:
        gradient_fn = jax.grad(inverse_link_function)
        gradient_fn(1.0)
    except Exception as e:
        raise ValueError(
            f"The `inverse_link_function` function cannot be differentiated. Error: {e}"
        ) from e



"""Collection of methods utilities."""

import difflib
import warnings
from typing import Any, List, Optional, Union

import jax
import jax.numpy as jnp
from numpy.typing import DTypeLike, NDArray

from . import utils
from .pytrees import FeaturePytree
from .tree_utils import get_valid_multitree, pytree_map_and_reduce


def error_invalid_entry(*pytree: Any):
    """
    Raise an error if any entry in the provided pytrees contains NaN or Infinite (Inf) values.

    Parameters
    ----------
    *pytree : Any
        Variable number of pytrees to be checked for invalid entries. A pytree is defined as a nested structure
        of lists, tuples, dictionaries, or other containers, with leaves that are arrays.

    Raises
    ------
    ValueError
        If any NaN or Inf values are found in the provided pytrees.
    """
    any_infs = pytree_map_and_reduce(
        jnp.any, any, jax.tree_util.tree_map(jnp.isinf, pytree)
    )
    any_nans = pytree_map_and_reduce(
        jnp.any, any, jax.tree_util.tree_map(jnp.isnan, pytree)
    )
    if any_infs and any_nans:
        raise ValueError("The provided trees contain Infs and Nans!")
    elif any_infs:
        raise ValueError("The provided trees contain Infs!")
    elif any_nans:
        raise ValueError("The provided trees contain Nans!")


def error_all_invalid(*pytree: Any):
    """
    Raise an error if all sample points across multiple pytrees are invalid.

    This function checks multiple pytrees with NDArrays as leaves to determine if all sample points are invalid.
    A sample point is considered invalid if it contains NaN or infinite values in at least one of the pytrees.
    The sample axis is the first dimension of the NDArray.

    Parameters
    ----------
    pytree :
        Variable number of pytrees to be evaluated. Each pytree is expected to have NDArrays as leaves with a
        consistent size for the first dimension (sample dimension).
        The function checks for the validity of sample points across these pytrees.

    Raises
    ------
    ValueError
        If all sample points across the provided pytrees are invalid (i.e., contain NaN or infinite values).
    """
    if all(~get_valid_multitree(*pytree)):
        raise ValueError("At least a NaN or an Inf at all sample points!")


def check_length(x: Any, expected_len: int, err_message: str):
    """
    Check if the provided object has a length of two.

    Parameters
    ----------
    x :
        Object to check the length of.
    expected_len :
        The expected length of the object.
    err_message :
        Error message to raise if the length is not two.

    Raises
    ------
    ValueError
        If the object does not have the specified length.
    """
    try:
        assert len(x) == expected_len
    except Exception:
        raise ValueError(err_message)


def convert_tree_leaves_to_jax_array(
    pytree: Any, err_message: str, data_type: Optional[DTypeLike] = None
):
    """
    Convert the leaves of a given pytree to JAX arrays with the specified data type.

    Parameters
    ----------
    pytree :
        Pytree with leaves that are array-like objects.
    err_message:
        The error message to raise if the leaves do not have the specified data type.
    data_type :
        Data type to convert the leaves to.

    Raises
    ------
    TypeError
        If conversion to JAX arrays fails due to incompatible types.

    Returns
    -------
    :
        A tree of the same structure as the original, with leaves converted
        to JAX arrays.
    """
    try:
        pytree = jax.tree_util.tree_map(
            lambda x: jnp.asarray(x, dtype=data_type), pytree
        )
    except (ValueError, TypeError):
        raise TypeError(err_message)
    return pytree


def check_tree_leaves_dimensionality(pytree: Any, expected_dim: int, err_message: str):
    """
    Check if the leaves of the pytree have the specified dimensionality.

    Parameters
    ----------
    pytree :
        Pytree to check the dimensionality of its leaves.
    expected_dim :
        Expected dimensionality of the leaves.
    err_message :
        Error message to raise if the dimensionality does not match.

    Raises
    ------
    ValueError
        If any leaf does not match the expected dimensionality.
    """
    if pytree_map_and_reduce(lambda x: x.ndim != expected_dim, any, pytree):
        raise ValueError(err_message)


def check_same_shape_on_axis(*arrays: NDArray, axis: int = 0, err_message: str):
    """
    Check if the arrays have the same shape along a specified axis.

    Parameters
    ----------
    *arrays :
        Arrays to check shape consistency of.
    axis :
        Axis along which to check the shape consistency.
    err_message :
        Error message to raise if the shapes are inconsistent.

    Raises
    ------
    ValueError
        If the arrays do not have the same shape along the specified axis.
    """
    if len(arrays) > 1:
        if any(arr.shape[axis] != arrays[0].shape[axis] for arr in arrays[1:]):
            raise ValueError(err_message)


def check_array_shape_match_tree(
    pytree: Any, array: NDArray, axis: int, err_message: str
):
    """
    Check if the shape of an array matches the shape of arrays in a pytree along a specified axis.

    Parameters
    ----------
    pytree :
        Pytree with arrays as leaves.
    array :
        Array to compare the shape with.
    axis :
        Axis along which to compare the shape.
    err_message :
        Error message to raise if the shapes do not match.

    Raises
    ------
    ValueError
        If the array's shape does not match the pytree leaves' shapes along the specified axis.
    """
    if pytree_map_and_reduce(
        lambda arr: arr.shape[axis] != array.shape[axis], any, pytree
    ):
        raise ValueError(err_message)


def array_axis_consistency(
    array_1: Union[FeaturePytree, jnp.ndarray, NDArray],
    array_2: Union[FeaturePytree, jnp.ndarray, NDArray],
    axis_1: int,
    axis_2: int,
):
    """
    Check if two arrays are consistent along specified axes.

    Parameters
    ----------
    array_1 :
        First array to check.
    array_2 :
        Second array to check.
    axis_1 :
        Axis to check in the first array.
    axis_2 :
        Axis to check in the second array.

    Returns
    -------
    bool
        True if inconsistent, otherwise False.
    """
    if array_1.shape[axis_1] != array_2.shape[axis_2]:
        return True
    else:
        return False


def check_tree_axis_consistency(
    pytree_1: Any,
    pytree_2: Any,
    axis_1: int,
    axis_2: int,
    err_message: str,
):
    """
    Check if two pytrees are consistent along specified axes for their respective leaves.

    Parameters
    ----------
    pytree_1 :
        First pytree to check.
    pytree_2 :
        Second pytree to check.
    axis_1 :
        Axis to check in the first pytree.
    axis_2 :
        Axis to check in the second pytree.
    err_message :
        Error message to raise if the pytrees' leaves are inconsistent along the given axes.

    Raises
    ------
    ValueError
        If the pytrees' leaves are inconsistent along the specified axes.
    """
    if pytree_map_and_reduce(
        lambda x, y: array_axis_consistency(x, y, axis_1, axis_2),
        any,
        pytree_1,
        pytree_2,
    ):
        raise ValueError(err_message)


def check_tree_structure(pytree_1: Any, pytree_2: Any, err_message: str):
    """Check if two pytrees have the same structure.

    Parameters
    ----------
    pytree_1 :
        First pytree to compare.
    pytree_2 :
        Second pytree to compare.
    err_message :
        Error message to raise if the structures of the pytrees do not match.

    Raises
    ------
    TypeError
        If the structures of the pytrees do not match.
    """
    if jax.tree_util.tree_structure(pytree_1) != jax.tree_util.tree_structure(pytree_2):
        raise TypeError(err_message)


def check_fraction_valid_samples(*pytree: Any, err_msg: str, warn_msg: str) -> None:
    """
    Check the fraction of entries that are not infinite or NaN.

    Parameters
    ----------
    *pytree :
        Trees containing arrays with the same sample axis.
    err_msg :
        The exception message.
    warn_msg :
        The warning message.

    Raises
    ------
    ValueError
        If all the samples contain invalid entries (either NaN or Inf).

    Warns
    -----
    UserWarning
        If more than 90% of the sample points contain NaNs or Infs.
    """
    valid = get_valid_multitree(pytree)
    if all(~valid):
        raise ValueError(err_msg)
    elif valid.mean() <= 0.1:
        warnings.warn(warn_msg, UserWarning)


def _warn_if_not_float64(feature_matrix: Any, message: str):
    """Warn if the feature matrix uses float32 precision."""
    all_float64 = pytree_map_and_reduce(
        lambda x: jnp.issubdtype(x.dtype, jnp.float64), all, feature_matrix
    )
    if not all_float64:
        warnings.warn(
            message,
            UserWarning,
        )


def _check_basis_matrix_shape(basis_matrix):
    basis_matrix = jnp.asarray(basis_matrix)
    if not utils.check_dimensionality(basis_matrix, 2):
        raise ValueError(
            "basis_matrix must be a 2 dimensional array! "
            f"{basis_matrix.ndim} dimensions provided instead."
        )
    if basis_matrix.shape[0] == 1:
        raise ValueError("`basis_matrix.shape[0]` should be at least 2!")
    return basis_matrix


def _check_non_empty_inputs(time_series, basis_matrix):
    utils.check_non_empty(basis_matrix, "basis_matrix")
    utils.check_non_empty(time_series, "time_series")


def _check_time_series_ndim(time_series, axis):
    if not utils.pytree_map_and_reduce(lambda x: x.ndim > axis, all, time_series):
        raise ValueError(
            "`time_series` should contain arrays of at least one-dimension. "
            "At least one 0-dimensional array provided."
        )


def _check_shift_causality_consistency(shift, predictor_causality):
    """Check shift causality consistency."""
    if shift and predictor_causality == "acausal":
        raise ValueError(
            "Cannot shift `predictor` when `predictor_causality` is `acausal`!"
        )


def _check_batch_size(batch_size, var_name):
    """Check if ``batch_size`` is a positive integer."""
    if batch_size is None:
        return
    elif not isinstance(batch_size, int) or batch_size < 1:
        raise ValueError(
            f"When provided ``{var_name}`` must be a strictly positive integer! "
            f"``{batch_size}`` provided instead."
        )


def _check_trials_longer_than_time_window(
    time_series: Any, window_size: int | Any, axis: int = 0
):
    """
    Check if the duration of each trial in the time series is at least as long as the window size.

    Parameters
    ----------
    time_series :
        A pytree of trial data.
    window_size :
        The size of the window to be used in convolution. Either an int or a pytree with the same
        struct as time_series.
    axis :
        The axis in the arrays representing the time dimension.

    Raises
    ------
    ValueError
        If any trial in the time series is shorter than the window size.
    """
    has_same_struct = jax.tree_util.tree_structure(
        time_series
    ) == jax.tree_util.tree_structure(window_size)
    if has_same_struct:
        insufficient_window_size = pytree_map_and_reduce(
            lambda x, w: x.shape[axis] < w, any, time_series, window_size
        )
    else:
        insufficient_window_size = pytree_map_and_reduce(
            lambda x: x.shape[axis] < window_size, any, time_series
        )
    # Check window size
    if insufficient_window_size:
        raise ValueError(
            "Insufficient trial duration. The number of time points in each trial must "
            "be greater or equal to the window size."
        )


def _check_batch_size_larger_than_convolution_window(
    batch_size: int | Any, window_size: int | Any
):
    """Check if the batch_size is larger than the window size."""
    has_same_struct = jax.tree_util.tree_structure(
        batch_size
    ) == jax.tree_util.tree_structure(window_size)
    if has_same_struct:
        insufficient_window_size = pytree_map_and_reduce(
            lambda x, w: x < w, any, batch_size, window_size
        )
    else:
        insufficient_window_size = pytree_map_and_reduce(
            lambda x: x < window_size, any, batch_size
        )
    if insufficient_window_size:
        bs = jax.tree_util.tree_leaves(batch_size)[0]
        ws = jax.tree_util.tree_leaves(window_size)[0]
        raise ValueError(
            "Batch size too small. Batch size must be larger than the convolution window size. "
            f"The provided batch size is ``{bs}``, while the window size for the convolution is ``{ws}``. "
            "Please increase the batch size."
        )


def _suggest_keys(
    unmatched_keys: List[str], valid_keys: List[str], cutoff: float = 0.6
):
    """
    Suggest the closest matching valid key for each unmatched key using fuzzy string matching.

    This function compares each unmatched key to a list of valid keys and returns a suggestion
    if a close match is found based on the similarity score.

    Parameters
    ----------
    unmatched_keys :
        Keys that were provided by the user but not found in the expected set.
    valid_keys :
        The list of valid/expected keys to compare against.
    cutoff :
        The minimum similarity ratio (between 0 and 1) required to consider a match.
        A higher value means stricter matching. Defaults to 0.6.

    Returns
    -------
    :
        A list of (provided_key, suggested_key) pairs. If no match is found,
        `suggested_key` will be `None`.

    Examples
    --------
    >>> _suggest_keys(["observaton_model"], ["observation_model", "regularization"])
    [('observaton_model', 'observation_model')]
    """
    key_paris = []  # format, (user_provided, similar key)
    for unmatched_key in unmatched_keys:
        suggestions = difflib.get_close_matches(
            unmatched_key, valid_keys, n=1, cutoff=cutoff
        )
        key_paris.append((unmatched_key, suggestions[0] if suggestions else None))
    return key_paris



"""Tests for making sure that solution reached for proximal operator is that the same as using another
method with just penalized loss."""

import jax
import numpy as np
import pytest
from scipy.optimize import minimize

import nemos as nmo

# Register every test here as solver-related
pytestmark = pytest.mark.solver_related


@pytest.mark.parametrize(
    "solver_names", [("GradientDescent", "ProximalGradient"), ("SVRG", "ProxSVRG")]
)
def test_unregularized_convergence(solver_names):
    """
    Assert that solution found when using GradientDescent vs ProximalGradient with an
    unregularized GLM is the same.
    """
    jax.config.update("jax_enable_x64", True)

    # generate toy data
    np.random.seed(111)
    # random design tensor. Shape (n_time_points, n_features).
    X = 0.5 * np.random.normal(size=(100, 5))

    # log-rates & weights, shape (1, ) and (n_features, ) respectively.
    b_true = np.zeros((1,))
    w_true = np.random.normal(size=(5,))

    # sparsify weights
    w_true[1:4] = 0.0

    # generate counts
    rate = jax.numpy.exp(jax.numpy.einsum("k,tk->t", w_true, X) + b_true)
    y = np.random.poisson(rate)

    # instantiate and fit unregularized GLM with GradientDescent
    model_GD = nmo.glm.GLM(solver_name=solver_names[0], solver_kwargs=dict(tol=10**-12))
    model_GD.fit(X, y)

    # instantiate and fit unregularized GLM with ProximalGradient
    model_PG = nmo.glm.GLM(solver_name=solver_names[1], solver_kwargs=dict(tol=10**-12))
    model_PG.fit(X, y)

    # assert weights are the same
    assert np.allclose(model_GD.coef_, model_PG.coef_)
    assert np.allclose(model_GD.intercept_, model_PG.intercept_)


@pytest.mark.parametrize(
    "solver_names", [("GradientDescent", "ProximalGradient"), ("SVRG", "ProxSVRG")]
)
def test_ridge_convergence(solver_names):
    """
    Assert that solution found when using GradientDescent vs ProximalGradient with an
    ridge GLM is the same.
    """
    jax.config.update("jax_enable_x64", True)
    # generate toy data
    np.random.seed(111)
    # random design tensor. Shape (n_time_points, n_features).
    X = 0.5 * np.random.normal(size=(100, 5))

    # log-rates & weights, shape (1, ) and (n_features, ) respectively.
    b_true = np.zeros((1,))
    w_true = np.random.normal(size=(5,))

    # sparsify weights
    w_true[1:4] = 0.0

    # generate counts
    rate = jax.numpy.exp(jax.numpy.einsum("k,tk->t", w_true, X) + b_true)
    y = np.random.poisson(rate)

    # instantiate and fit ridge GLM with GradientDescent
    model_GD = nmo.glm.GLM(
        regularizer_strength=1.0,
        regularizer="Ridge",
        solver_kwargs=dict(tol=10**-12),
        solver_name=solver_names[0],
    )
    model_GD.fit(X, y)

    # instantiate and fit ridge GLM with ProximalGradient
    model_PG = nmo.glm.GLM(
        regularizer_strength=1.0,
        regularizer="Ridge",
        solver_name=solver_names[1],
        solver_kwargs=dict(tol=10**-12),
    )
    model_PG.fit(X, y)

    # assert weights are the same
    assert np.allclose(model_GD.coef_, model_PG.coef_)
    assert np.allclose(model_GD.intercept_, model_PG.intercept_)


@pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
def test_lasso_convergence(solver_name):
    """
    Assert that solution found when using ProximalGradient versus Nelder-Mead method using
    lasso GLM is the same.
    """
    jax.config.update("jax_enable_x64", True)
    # generate toy data
    num_samples, num_features, num_groups = 1000, 1, 3
    X = np.random.normal(size=(num_samples, num_features))  # design matrix
    w = [0.5]  # define some weights
    y = np.random.poisson(np.exp(X.dot(w)))  # observed counts

    # instantiate and fit GLM with ProximalGradient
    model_PG = nmo.glm.GLM(
        regularizer="Lasso",
        regularizer_strength=1.0,
        solver_name=solver_name,
        solver_kwargs=dict(tol=10**-12),
    )
    model_PG.regularizer_strength = 0.1
    model_PG.fit(X, y)

    # use the penalized loss function to solve optimization via Nelder-Mead
    penalized_loss = lambda p, x, y: model_PG.regularizer.penalized_loss(
        model_PG._predict_and_compute_loss, model_PG.regularizer_strength
    )(
        (
            p[1:],
            p[0].reshape(
                1,
            ),
        ),
        x,
        y,
    )
    res = minimize(
        penalized_loss, [0] + w, args=(X, y), method="Nelder-Mead", tol=10**-12
    )

    # assert weights are the same
    assert np.allclose(res.x[1:], model_PG.coef_)
    assert np.allclose(res.x[:1], model_PG.intercept_)


@pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
def test_group_lasso_convergence(solver_name):
    """
    Assert that solution found when using ProximalGradient versus Nelder-Mead method using
    group lasso GLM is the same.
    """
    jax.config.update("jax_enable_x64", True)
    # generate toy data
    num_samples, num_features, num_groups = 1000, 3, 2
    X = np.random.normal(size=(num_samples, num_features))  # design matrix
    w = [-0.5, 0.25, 0.5]  # define some weights
    y = np.random.poisson(np.exp(X.dot(w)))  # observed counts

    mask = np.zeros((num_groups, num_features))
    mask[0] = [1, 1, 0]  # Group 0 includes features 0 and 1
    mask[1] = [0, 0, 1]  # Group 1 includes features 1

    # instantiate and fit GLM with ProximalGradient
    model_PG = nmo.glm.GLM(
        regularizer=nmo.regularizer.GroupLasso(mask=mask),
        solver_kwargs=dict(tol=10**-14, maxiter=100_000),
        regularizer_strength=0.2,
        solver_name=solver_name,
    )
    model_PG.fit(X, y)

    # use the penalized loss function to solve optimization via Nelder-Mead
    penalized_loss = lambda p, x, y: model_PG.regularizer.penalized_loss(
        model_PG._predict_and_compute_loss, model_PG.regularizer_strength
    )(
        (
            p[1:],
            p[0].reshape(
                1,
            ),
        ),
        x,
        y,
    )

    res = minimize(
        penalized_loss,
        [0] + w,
        args=(X, y),
        method="Nelder-Mead",
        tol=10**-12,
        options=dict(maxiter=1000),
    )

    # assert weights are the same
    assert np.allclose(res.x[1:], model_PG.coef_)
    assert np.allclose(res.x[:1], model_PG.intercept_)



import copy
import warnings
from contextlib import nullcontext as does_not_raise

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import statsmodels.api as sm
from scipy.optimize import minimize
from sklearn.linear_model import GammaRegressor, PoissonRegressor
from statsmodels.tools.sm_exceptions import DomainWarning

import nemos as nmo

# Register every test here as solver-related
pytestmark = pytest.mark.solver_related


@pytest.mark.parametrize(
    "reg_str, reg_type",
    [
        ("UnRegularized", nmo.regularizer.UnRegularized),
        (None, nmo.regularizer.UnRegularized),
        ("Ridge", nmo.regularizer.Ridge),
        ("Lasso", nmo.regularizer.Lasso),
        ("GroupLasso", nmo.regularizer.GroupLasso),
        ("ElasticNet", nmo.regularizer.ElasticNet),
        ("not_valid", None),
        ("nemos.regularizer.UnRegularized", nmo.regularizer.UnRegularized),
        ("nemos.regularizer.Ridge", nmo.regularizer.Ridge),
        ("nemos.regularizer.Lasso", nmo.regularizer.Lasso),
        ("nemos.regularizer.GroupLasso", nmo.regularizer.GroupLasso),
    ],
)
def test_regularizer_builder(reg_str, reg_type):
    """Test building a regularizer from a string"""
    valid_regularizers = nmo._regularizer_builder.AVAILABLE_REGULARIZERS
    raise_exception = reg_str is not None and not (
        reg_str in valid_regularizers
        or any(reg_str == f"nemos.regularizer.{name}" for name in valid_regularizers)
    )
    if raise_exception:
        with pytest.raises(ValueError, match=f"Unknown regularizer: {reg_str}. "):
            nmo._regularizer_builder.instantiate_regularizer(reg_str)
    else:
        # build a regularizer by string
        regularizer = nmo._regularizer_builder.instantiate_regularizer(reg_str)
        # assert correct type of regularizer is instantiated
        assert isinstance(regularizer, reg_type)
        # create a regularizer of that type
        regularizer2 = reg_type()
        # assert that they have the same attributes
        assert regularizer.__dict__ == regularizer2.__dict__


@pytest.mark.parametrize(
    "expected, reg",
    [
        ("UnRegularized()", nmo.regularizer.UnRegularized()),
        ("Ridge()", nmo.regularizer.Ridge()),
        ("Lasso()", nmo.regularizer.Lasso()),
        ("GroupLasso()", nmo.regularizer.GroupLasso(mask=np.eye(4))),
        ("ElasticNet()", nmo.regularizer.ElasticNet()),
    ],
)
def test_regularizer_repr(reg, expected):
    assert repr(reg) == expected


def test_regularizer_available():
    for regularizer in nmo._regularizer_builder.AVAILABLE_REGULARIZERS:
        reg = nmo._regularizer_builder.instantiate_regularizer(regularizer)
        assert reg.__class__.__name__ == regularizer


@pytest.mark.parametrize(
    "regularizer_strength",
    [0.001, 1.0, "bah"],
)
@pytest.mark.parametrize(
    "reg_type",
    [
        nmo.regularizer.Ridge,
        nmo.regularizer.Lasso,
        nmo.regularizer.GroupLasso,
        nmo.regularizer.ElasticNet,
    ],
)
def test_regularizer(regularizer_strength, reg_type):
    if not isinstance(regularizer_strength, float):
        with pytest.raises(
            ValueError,
            match=r"Could not convert the (regularizer strength|regularizer strength and regularizer ratio): {regularizer_strength} "
            r"to a (float|tuple of floats).".format(
                regularizer_strength=regularizer_strength
            ),
        ):
            nmo.glm.GLM(
                regularizer=reg_type(), regularizer_strength=regularizer_strength
            )
    else:
        nmo.glm.GLM(regularizer=reg_type(), regularizer_strength=regularizer_strength)


@pytest.mark.parametrize(
    "regularizer_strength",
    [0.001, 1.0, "bah"],
)
@pytest.mark.parametrize(
    "regularizer",
    [
        nmo.regularizer.Ridge(),
        nmo.regularizer.Lasso(),
        nmo.regularizer.GroupLasso(),
        nmo.regularizer.ElasticNet(),
    ],
)
def test_regularizer_setter(regularizer_strength, regularizer):
    if not isinstance(regularizer_strength, float):
        with pytest.raises(
            ValueError,
            match=r"Could not convert the (regularizer strength|regularizer strength and regularizer ratio): {regularizer_strength} "
            r"to a (float|tuple of floats).".format(
                regularizer_strength=regularizer_strength
            ),
        ):
            nmo.glm.GLM(
                regularizer=regularizer, regularizer_strength=regularizer_strength
            )
    else:
        nmo.glm.GLM(regularizer=regularizer, regularizer_strength=regularizer_strength)


@pytest.mark.parametrize(
    "regularizer",
    [
        nmo.regularizer.UnRegularized(),
        nmo.regularizer.Ridge(),
        nmo.regularizer.Lasso(),
        nmo.regularizer.GroupLasso(mask=np.array([[1.0]])),
        nmo.regularizer.ElasticNet(),
    ],
)
def test_get_only_allowed_solvers(regularizer):
    # the error raised by property changed in python 3.11
    with pytest.raises(
        AttributeError,
        match="property 'allowed_solvers' of '.+' object has no setter|can't set attribute",
    ):
        regularizer.allowed_solvers = []


@pytest.mark.parametrize(
    "regularizer",
    [
        nmo.regularizer.UnRegularized(),
        nmo.regularizer.Ridge(),
        nmo.regularizer.Lasso(),
        nmo.regularizer.GroupLasso(mask=np.array([[1.0]])),
        nmo.regularizer.ElasticNet(),
    ],
)
def test_item_assignment_allowed_solvers(regularizer):
    with pytest.raises(
        TypeError, match="'tuple' object does not support item assignment"
    ):
        regularizer.allowed_solvers[0] = "my-favourite-solver"


@pytest.mark.parametrize(
    "regularizer, regularizer_strength",
    [
        ("Ridge", 2.0),
        ("Lasso", 2.0),
        ("GroupLasso", 2.0),
        ("ElasticNet", (2.0, 0.7)),
    ],
)
def test_set_params_order_change_regularizer(regularizer, regularizer_strength):
    """Test that set_params() when changing regularizer and regularizer_strength regardless of order."""
    # start with unregularized
    model = nmo.glm.GLM()
    assert model.regularizer_strength is None
    assert isinstance(model.regularizer, nmo.regularizer.UnRegularized)

    # set regularizer first
    model.set_params(regularizer=regularizer, regularizer_strength=regularizer_strength)
    assert model.regularizer_strength == regularizer_strength
    assert model.regularizer.__class__.__name__ == regularizer

    # set regularizer_strength first
    model.set_params(regularizer="UnRegularized")
    assert model.regularizer_strength is None

    model.set_params(regularizer_strength=regularizer_strength, regularizer=regularizer)
    assert model.regularizer_strength == regularizer_strength
    assert model.regularizer.__class__.__name__ == regularizer


@pytest.mark.parametrize(
    "regularizer, regularizer_strength",
    [
        ("Ridge", 2.0),
        ("Lasso", 2.0),
        ("GroupLasso", 2.0),
        ("ElasticNet", (2.0, 0.7)),
        ("UnRegularized", None),
    ],
)
@pytest.mark.parametrize(
    "regularizer2, regularizer2_default",
    [
        ("Ridge", 1.0),
        ("Lasso", 1.0),
        ("GroupLasso", 1.0),
        ("ElasticNet", (1.0, 0.5)),
        ("UnRegularized", None),
    ],
)
def test_change_regularizer_reset_strength(
    regularizer,
    regularizer_strength,
    regularizer2,
    regularizer2_default,
):
    """Test that set_params() when changing regularizer and regularizer_strength regardless of order."""
    model = nmo.glm.GLM(
        regularizer=regularizer, regularizer_strength=regularizer_strength
    )
    assert model.regularizer_strength == regularizer_strength
    assert model.regularizer.__class__.__name__ == regularizer

    # check that regularizer_strength is reset when changing regularizer
    model.set_params(regularizer=regularizer2)
    assert model.regularizer_strength == regularizer2_default

    # make sure there is no conflict when setting back
    model.set_params(regularizer=regularizer, regularizer_strength=regularizer_strength)
    assert model.regularizer_strength == regularizer_strength


class TestUnRegularized:
    cls = nmo.regularizer.UnRegularized

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_init_solver_name(self, solver_name):
        """Test UnRegularized acceptable solvers."""
        acceptable_solvers = [
            "GradientDescent",
            "BFGS",
            "LBFGSB",
            "NonlinearCG",
            "ProximalGradient",
            "SVRG",
            "ProxSVRG",
        ]

        raise_exception = solver_name not in acceptable_solvers
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                nmo.glm.GLM(regularizer=self.cls(), solver_name=solver_name)
        else:
            nmo.glm.GLM(regularizer=self.cls(), solver_name=solver_name)

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_set_solver_name_allowed(self, solver_name):
        """Test UnRegularized acceptable solvers."""
        acceptable_solvers = [
            "GradientDescent",
            "BFGS",
            "LBFGS",
            "NonlinearCG",
            "ProximalGradient",
            "SVRG",
            "ProxSVRG",
        ]
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer)
        raise_exception = solver_name not in acceptable_solvers
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                model.set_params(solver_name=solver_name)
        else:
            model.set_params(solver_name=solver_name)

    def test_regularizer_strength_none(self):
        """Add test to assert that regularizer strength of UnRegularized model should be `None`"""
        # unregularized should be None
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer)

        assert model.regularizer_strength is None

        # changing to ridge, lasso, or grouplasso should set to 1.0
        model.regularizer = "Lasso"

        assert model.regularizer_strength == 1.0
        model.regularizer = regularizer
        assert model.regularizer_strength == None

    def test_get_params(self):
        """Test get_params() returns expected values."""
        regularizer = self.cls()

        assert regularizer.get_params() == {}

    @pytest.mark.parametrize(
        "solver_name", ["GradientDescent", "BFGS", "SVRG", "ProxSVRG"]
    )
    @pytest.mark.parametrize("solver_kwargs", [{"tol": 10**-10}, {"tols": 10**-10}])
    def test_init_solver_kwargs(self, solver_name, solver_kwargs):
        """Test RidgeSolver acceptable kwargs."""
        regularizer = self.cls()
        raise_exception = "tols" in list(solver_kwargs.keys())
        if raise_exception:
            with pytest.raises(
                NameError, match="kwargs {'tols'} in solver_kwargs not a kwarg"
            ):
                nmo.glm.GLM(
                    regularizer=regularizer,
                    solver_name=solver_name,
                    solver_kwargs=solver_kwargs,
                )
        else:
            nmo.glm.GLM(
                regularizer=regularizer,
                solver_name=solver_name,
                solver_kwargs=solver_kwargs,
            )

    @pytest.mark.parametrize("loss", [lambda a, b, c: 0, 1, None, {}])
    def test_loss_is_callable(self, loss):
        """Test Unregularized callable loss."""
        raise_exception = not callable(loss)
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer)
        model._predict_and_compute_loss = loss
        if raise_exception:
            with pytest.raises(TypeError, match="The `loss` must be a Callable"):
                nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")
        else:
            nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")

    @pytest.mark.parametrize(
        "solver_name",
        ["GradientDescent", "BFGS", "ProximalGradient", "SVRG", "ProxSVRG"],
    )
    def test_run_solver(self, solver_name, poissonGLM_model_instantiation):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # set regularizer and solver name
        model.set_params(regularizer=self.cls())
        model.solver_name = solver_name
        model.instantiate_solver()
        model.solver_run((true_params[0] * 0.0, true_params[1]), X, y)

    @pytest.mark.parametrize(
        "solver_name",
        ["GradientDescent", "BFGS", "ProximalGradient", "SVRG", "ProxSVRG"],
    )
    def test_run_solver_tree(self, solver_name, poissonGLM_model_instantiation_pytree):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation_pytree

        # set regularizer and solver name
        model.set_params(regularizer=self.cls())
        model.solver_name = solver_name
        model.instantiate_solver()
        model.solver_run(
            (jax.tree_util.tree_map(jnp.zeros_like, true_params[0]), true_params[1]),
            X.data,
            y,
        )

    @pytest.mark.parametrize("solver_name", ["GradientDescent", "SVRG"])
    def test_solver_output_match(self, poissonGLM_model_instantiation, solver_name):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        # set model params
        model.set_params(regularizer=self.cls())
        model.solver_name = solver_name
        model.solver_kwargs = {"tol": 10**-12}
        model.instantiate_solver()

        # update solver name
        model_bfgs = copy.deepcopy(model)
        model_bfgs.solver_name = "BFGS"
        model_bfgs.instantiate_solver()
        weights_gd, intercepts_gd = model.solver_run(
            (true_params[0] * 0.0, true_params[1]), X, y
        )[0]
        weights_bfgs, intercepts_bfgs = model_bfgs.solver_run(
            (true_params[0] * 0.0, true_params[1]), X, y
        )[0]

        match_weights = np.allclose(weights_gd, weights_bfgs)
        match_intercepts = np.allclose(intercepts_gd, intercepts_bfgs)

        if (not match_weights) or (not match_intercepts):
            raise ValueError(
                "Convex estimators should converge to the same numerical value."
            )

    @pytest.mark.parametrize("solver_name", ["GradientDescent", "SVRG"])
    def test_solver_match_sklearn(self, poissonGLM_model_instantiation, solver_name):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        model.set_params(regularizer=self.cls())
        model.solver_name = solver_name
        model.solver_kwargs = {"tol": 10**-12}
        model.instantiate_solver()
        weights_bfgs, intercepts_bfgs = model.solver_run(
            (true_params[0] * 0.0, true_params[1]), X, y
        )[0]
        model_skl = PoissonRegressor(fit_intercept=True, tol=10**-12, alpha=0.0)
        model_skl.fit(X, y)

        match_weights = np.allclose(model_skl.coef_, weights_bfgs)
        match_intercepts = np.allclose(model_skl.intercept_, intercepts_bfgs)
        if (not match_weights) or (not match_intercepts):
            raise ValueError("UnRegularized GLM estimate does not match sklearn!")

    @pytest.mark.parametrize("solver_name", ["GradientDescent", "SVRG"])
    def test_solver_match_sklearn_gamma(
        self, gammaGLM_model_instantiation, solver_name
    ):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = gammaGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        model.inverse_link_function = jnp.exp
        model.set_params(regularizer=self.cls())
        model.solver_name = solver_name
        model.solver_kwargs = {"tol": 10**-12}
        model.instantiate_solver()
        weights_bfgs, intercepts_bfgs = model.solver_run(
            (true_params[0] * 0.0, true_params[1]), X, y
        )[0]
        model_skl = GammaRegressor(fit_intercept=True, tol=10**-12, alpha=0.0)
        model_skl.fit(X, y)

        match_weights = np.allclose(model_skl.coef_, weights_bfgs)
        match_intercepts = np.allclose(model_skl.intercept_, intercepts_bfgs)
        if (not match_weights) or (not match_intercepts):
            raise ValueError("Unregularized GLM estimate does not match sklearn!")

    @pytest.mark.parametrize(
        "inv_link_jax, link_sm",
        [
            (jnp.exp, sm.families.links.Log()),
            (lambda x: 1 / x, sm.families.links.InversePower()),
        ],
    )
    # @pytest.mark.parametrize("solver_name", ["LBFGS", "GradientDescent", "SVRG"])
    @pytest.mark.parametrize("solver_name", ["LBFGS", "SVRG"])
    def test_solver_match_statsmodels_gamma(
        self, inv_link_jax, link_sm, gammaGLM_model_instantiation, solver_name
    ):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = gammaGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        model.inverse_link_function = inv_link_jax
        model.set_params(regularizer=self.cls())
        model.solver_name = solver_name
        model.solver_kwargs = {"tol": 10**-13}
        model.instantiate_solver()
        weights_bfgs, intercepts_bfgs = model.solver_run(
            model._initialize_parameters(X, y), X, y
        )[0]
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="The InversePower link function does "
            )
            model_sm = sm.GLM(
                endog=y, exog=sm.add_constant(X), family=sm.families.Gamma(link=link_sm)
            )

        res_sm = model_sm.fit(cnvrg_tol=10**-12)

        match_weights = np.allclose(res_sm.params[1:], weights_bfgs)
        match_intercepts = np.allclose(res_sm.params[:1], intercepts_bfgs)
        if (not match_weights) or (not match_intercepts):
            raise ValueError("Unregularized GLM estimate does not match statsmodels!")

    @pytest.mark.parametrize(
        "inv_link_jax, link_sm",
        [
            (jnp.exp, sm.families.links.Log()),
        ],
    )
    @pytest.mark.parametrize("solver_name", ["LBFGS", "SVRG"])
    def test_solver_match_statsmodels_negative_binomial(
        self,
        inv_link_jax,
        link_sm,
        negativeBinomialGLM_model_instantiation,
        solver_name,
    ):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = negativeBinomialGLM_model_instantiation
        y = y.astype(
            float
        )  # needed since solver.run is called directly, nemos converts.
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        model.observation_model.inverse_link_function = inv_link_jax
        model.set_params(regularizer=self.cls())
        model.solver_name = solver_name
        model.solver_kwargs = {"tol": 10**-13}
        model.instantiate_solver()
        weights_bfgs, intercepts_bfgs = model.solver_run(
            model._initialize_parameters(X, y), X, y
        )[0]
        model_sm = sm.GLM(
            endog=y,
            exog=sm.add_constant(X),
            family=sm.families.NegativeBinomial(
                link=link_sm, alpha=model.observation_model.scale
            ),
        )

        res_sm = model_sm.fit(cnvrg_tol=10**-12)

        match_weights = np.allclose(res_sm.params[1:], weights_bfgs, atol=10**-6)
        match_intercepts = np.allclose(res_sm.params[:1], intercepts_bfgs, atol=10**-6)
        if (not match_weights) or (not match_intercepts):
            raise ValueError(
                "Unregularized GLM estimate does not match statsmodels!\n"
                f"Intercept difference is: {res_sm.params[:1] - intercepts_bfgs}\n"
                f"Coefficient difference is: {res_sm.params[1:] - weights_bfgs}"
            )

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "LBFGS",
            "NonlinearCG",
            "ProximalGradient",
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_solver_combination(self, solver_name, poissonGLM_model_instantiation):
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        model.set_params(regularizer=self.cls())
        model.solver_name = solver_name
        model.fit(X, y)


class TestRidge:
    cls = nmo.regularizer.Ridge

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_init_solver_name(self, solver_name):
        """Test RidgeSolver acceptable solvers."""
        acceptable_solvers = [
            "GradientDescent",
            "BFGS",
            "LBFGS",
            "NonlinearCG",
            "ProximalGradient",
            "SVRG",
            "ProxSVRG",
        ]
        raise_exception = solver_name not in acceptable_solvers
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                nmo.glm.GLM(
                    regularizer=self.cls(),
                    solver_name=solver_name,
                    regularizer_strength=1.0,
                )
        else:
            nmo.glm.GLM(
                regularizer=self.cls(),
                solver_name=solver_name,
                regularizer_strength=1.0,
            )

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_set_solver_name_allowed(self, solver_name):
        """Test RidgeSolver acceptable solvers."""
        acceptable_solvers = [
            "GradientDescent",
            "BFGS",
            "LBFGS",
            "LBFGSB",
            "NonlinearCG",
            "ProximalGradient",
            "SVRG",
            "ProxSVRG",
        ]
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=1.0)
        raise_exception = solver_name not in acceptable_solvers
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                model.set_params(solver_name=solver_name)
        else:
            model.set_params(solver_name=solver_name)

    @pytest.mark.parametrize("solver_name", ["GradientDescent", "BFGS", "SVRG"])
    @pytest.mark.parametrize("solver_kwargs", [{"tol": 10**-10}, {"tols": 10**-10}])
    def test_init_solver_kwargs(self, solver_name, solver_kwargs):
        """Test Ridge acceptable kwargs."""
        regularizer = self.cls()
        raise_exception = "tols" in list(solver_kwargs.keys())
        if raise_exception:
            with pytest.raises(
                NameError, match="kwargs {'tols'} in solver_kwargs not a kwarg"
            ):
                nmo.glm.GLM(
                    regularizer=regularizer,
                    solver_name=solver_name,
                    solver_kwargs=solver_kwargs,
                    regularizer_strength=1.0,
                )
        else:
            nmo.glm.GLM(
                regularizer=regularizer,
                solver_name=solver_name,
                solver_kwargs=solver_kwargs,
                regularizer_strength=1.0,
            )

    def test_regularizer_strength_none(self):
        """Assert regularizer strength handled appropriately."""
        # if no strength given, should set to 1.0
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer)

        assert model.regularizer_strength == 1.0
        # if changed to regularized, is set to None.
        model.regularizer = "UnRegularized"
        assert model.regularizer_strength is None

        # if changed back, should set to 1.0
        model.regularizer = "Ridge"

        assert model.regularizer_strength == 1.0

    def test_get_params(self):
        """Test get_params() returns expected values."""
        regularizer = self.cls()

        assert regularizer.get_params() == {}

    @pytest.mark.parametrize("loss", [lambda a, b, c: 0, 1, None, {}])
    def test_loss_is_callable(self, loss):
        """Test Ridge callable loss."""
        raise_exception = not callable(loss)
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=1.0)
        model._predict_and_compute_loss = loss
        if raise_exception:
            with pytest.raises(TypeError, match="The `loss` must be a Callable"):
                nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")
        else:
            nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")

    @pytest.mark.parametrize(
        "solver_name",
        ["GradientDescent", "BFGS", "ProximalGradient", "SVRG", "ProxSVRG"],
    )
    def test_run_solver(self, solver_name, poissonGLM_model_instantiation):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # set regularizer and solver name
        model.set_params(regularizer=self.cls(), regularizer_strength=1.0)
        model.solver_name = solver_name
        runner = model.instantiate_solver().solver_run
        runner((true_params[0] * 0.0, true_params[1]), X, y)

    @pytest.mark.parametrize(
        "solver_name",
        ["GradientDescent", "BFGS", "ProximalGradient", "SVRG", "ProxSVRG"],
    )
    def test_run_solver_tree(self, solver_name, poissonGLM_model_instantiation_pytree):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation_pytree

        # set regularizer and solver name
        model.set_params(regularizer=self.cls(), regularizer_strength=1.0)
        model.solver_name = solver_name
        runner = model.instantiate_solver().solver_run
        runner(
            (jax.tree_util.tree_map(jnp.zeros_like, true_params[0]), true_params[1]),
            X.data,
            y,
        )

    @pytest.mark.parametrize("solver_name", ["GradientDescent", "SVRG"])
    def test_solver_output_match(self, poissonGLM_model_instantiation, solver_name):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64

        # set model params
        model.set_params(regularizer=self.cls(), regularizer_strength=1.0)
        model.solver_name = solver_name
        model.solver_kwargs = {"tol": 10**-12}

        model_bfgs = copy.deepcopy(model)
        model_bfgs.solver_name = "BFGS"

        runner_gd = model.instantiate_solver().solver_run
        runner_bfgs = model_bfgs.instantiate_solver().solver_run

        weights_gd, intercepts_gd = runner_gd(
            (true_params[0] * 0.0, true_params[1]), X, y
        )[0]
        weights_bfgs, intercepts_bfgs = runner_bfgs(
            (true_params[0] * 0.0, true_params[1]), X, y
        )[0]

        match_weights = np.allclose(weights_gd, weights_bfgs)
        match_intercepts = np.allclose(intercepts_gd, intercepts_bfgs)

        if (not match_weights) or (not match_intercepts):
            raise ValueError(
                "Convex estimators should converge to the same numerical value."
            )

    def test_solver_match_sklearn(self, poissonGLM_model_instantiation):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        model.set_params(regularizer=self.cls(), regularizer_strength=1.0)
        model.solver_kwargs = {"tol": 10**-12}
        model.solver_name = "BFGS"

        runner_bfgs = model.instantiate_solver().solver_run
        weights_bfgs, intercepts_bfgs = runner_bfgs(
            (true_params[0] * 0.0, true_params[1]), X, y
        )[0]
        model_skl = PoissonRegressor(
            fit_intercept=True,
            tol=10**-12,
            alpha=model.regularizer_strength,
        )
        model_skl.fit(X, y)

        match_weights = np.allclose(model_skl.coef_, weights_bfgs)
        match_intercepts = np.allclose(model_skl.intercept_, intercepts_bfgs)
        if (not match_weights) or (not match_intercepts):
            raise ValueError("Ridge GLM solver estimate does not match sklearn!")

    def test_solver_match_sklearn_gamma(self, gammaGLM_model_instantiation):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = gammaGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        model.inverse_link_function = jnp.exp
        model.set_params(regularizer=self.cls(), regularizer_strength=1.0)
        model.solver_kwargs = {"tol": 10**-12}
        model.regularizer_strength = 0.1
        model.solver_name = "BFGS"
        runner_bfgs = model.instantiate_solver().solver_run
        weights_bfgs, intercepts_bfgs = runner_bfgs(
            (true_params[0] * 0.0, true_params[1]), X, y
        )[0]
        model_skl = GammaRegressor(
            fit_intercept=True,
            tol=10**-12,
            alpha=model.regularizer_strength,
        )
        model_skl.fit(X, y)

        match_weights = np.allclose(model_skl.coef_, weights_bfgs)
        match_intercepts = np.allclose(model_skl.intercept_, intercepts_bfgs)
        if (not match_weights) or (not match_intercepts):
            raise ValueError("Ridge GLM estimate does not match sklearn!")

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "LBFGS",
            "NonlinearCG",
            "ProximalGradient",
        ],
    )
    def test_solver_combination(self, solver_name, poissonGLM_model_instantiation):
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        model.set_params(regularizer=self.cls(), regularizer_strength=1.0)
        model.solver_name = solver_name
        model.fit(X, y)


class TestLasso:
    cls = nmo.regularizer.Lasso

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_init_solver_name(self, solver_name):
        """Test Lasso acceptable solvers."""
        acceptable_solvers = [
            "ProximalGradient",
            "ProxSVRG",
        ]
        raise_exception = solver_name not in acceptable_solvers
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                nmo.glm.GLM(
                    regularizer=self.cls(),
                    solver_name=solver_name,
                    regularizer_strength=1,
                )
        else:
            nmo.glm.GLM(
                regularizer=self.cls(), solver_name=solver_name, regularizer_strength=1
            )

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_set_solver_name_allowed(self, solver_name):
        """Test Lasso acceptable solvers."""
        acceptable_solvers = [
            "ProximalGradient",
            "ProxSVRG",
        ]
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=1)
        raise_exception = solver_name not in acceptable_solvers
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                model.set_params(solver_name=solver_name)
        else:
            model.set_params(solver_name=solver_name)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    @pytest.mark.parametrize("solver_kwargs", [{"tol": 10**-10}, {"tols": 10**-10}])
    def test_init_solver_kwargs(self, solver_kwargs, solver_name):
        """Test LassoSolver acceptable kwargs."""
        regularizer = self.cls()
        raise_exception = "tols" in list(solver_kwargs.keys())
        if raise_exception:
            with pytest.raises(
                NameError, match="kwargs {'tols'} in solver_kwargs not a kwarg"
            ):
                nmo.glm.GLM(
                    regularizer=regularizer,
                    solver_name=solver_name,
                    solver_kwargs=solver_kwargs,
                    regularizer_strength=1.0,
                )
        else:
            nmo.glm.GLM(
                regularizer=regularizer,
                solver_name=solver_name,
                solver_kwargs=solver_kwargs,
                regularizer_strength=1.0,
            )

    def test_regularizer_strength_none(self):
        """Assert regularizer strength handled appropriately."""
        # if no strength given, should set to 1.0
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer)

        assert model.regularizer_strength == 1.0

        # if changed to regularized, should go to None
        model.set_params(regularizer="UnRegularized", regularizer_strength=None)
        assert model.regularizer_strength is None

        # if changed back, should set to 1.0
        model.regularizer = regularizer

        assert model.regularizer_strength == 1.0

    def test_get_params(self):
        """Test get_params() returns expected values."""
        regularizer = self.cls()

        assert regularizer.get_params() == {}

    @pytest.mark.parametrize("loss", [lambda a, b, c: 0, 1, None, {}])
    def test_loss_callable(self, loss):
        """Test that the loss function is a callable"""
        raise_exception = not callable(loss)
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=1)
        model._predict_and_compute_loss = loss
        if raise_exception:
            with pytest.raises(TypeError, match="The `loss` must be a Callable"):
                nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")
        else:
            nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_run_solver(self, solver_name, poissonGLM_model_instantiation):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        model.set_params(regularizer=self.cls(), regularizer_strength=1)
        model.solver_name = solver_name
        runner = model.instantiate_solver().solver_run
        runner((true_params[0] * 0.0, true_params[1]), X, y)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_run_solver_tree(self, solver_name, poissonGLM_model_instantiation_pytree):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation_pytree

        # set regularizer and solver name
        model.set_params(regularizer=self.cls(), regularizer_strength=1)
        model.solver_name = solver_name
        runner = model.instantiate_solver().solver_run
        runner(
            (jax.tree_util.tree_map(jnp.zeros_like, true_params[0]), true_params[1]),
            X.data,
            y,
        )

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_solver_match_statsmodels(
        self, solver_name, poissonGLM_model_instantiation
    ):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        model.set_params(regularizer=self.cls(), regularizer_strength=1)
        model.solver_name = solver_name
        model.solver_kwargs = {"tol": 10**-12}

        runner = model.instantiate_solver().solver_run
        weights, intercepts = runner((true_params[0] * 0.0, true_params[1]), X, y)[0]

        # instantiate the glm with statsmodels
        glm_sm = sm.GLM(endog=y, exog=sm.add_constant(X), family=sm.families.Poisson())

        # regularize everything except intercept
        alpha_sm = np.ones(X.shape[1] + 1) * model.regularizer_strength
        alpha_sm[0] = 0

        # pure lasso = elastic net with L1 weight = 1
        res_sm = glm_sm.fit_regularized(
            method="elastic_net", alpha=alpha_sm, L1_wt=1.0, cnvrg_tol=10**-12
        )
        # compare params
        sm_params = res_sm.params
        glm_params = jnp.hstack((intercepts, weights.flatten()))
        match_weights = np.allclose(sm_params, glm_params)
        if not match_weights:
            raise ValueError("Lasso GLM solver estimate does not match statsmodels!")

    def test_lasso_pytree(self, poissonGLM_model_instantiation_pytree):
        """Check pytree X can be fit."""
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation_pytree
        model.set_params(regularizer=nmo.regularizer.Lasso(), regularizer_strength=1.0)
        model.solver_name = "ProximalGradient"
        model.fit(X, y)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    @pytest.mark.parametrize("reg_str", [0.001, 0.01, 0.1, 1, 10])
    def test_lasso_pytree_match(
        self,
        reg_str,
        solver_name,
        poissonGLM_model_instantiation_pytree,
        poissonGLM_model_instantiation,
    ):
        """Check pytree and array find same solution."""
        jax.config.update("jax_enable_x64", True)
        X, _, model, _, _ = poissonGLM_model_instantiation_pytree
        X_array, y, model_array, _, _ = poissonGLM_model_instantiation

        model.set_params(
            regularizer=nmo.regularizer.Lasso(), regularizer_strength=reg_str
        )
        model_array.set_params(
            regularizer=nmo.regularizer.Lasso(), regularizer_strength=reg_str
        )
        model.solver_name = solver_name
        model_array.solver_name = solver_name
        model.fit(X, y)
        model_array.fit(X_array, y)
        assert np.allclose(
            np.hstack(jax.tree_util.tree_leaves(model.coef_)), model_array.coef_
        )

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_solver_combination(self, solver_name, poissonGLM_model_instantiation):
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        model.set_params(regularizer=self.cls(), regularizer_strength=1.0)
        model.solver_name = solver_name
        model.fit(X, y)


class TestElasticNet:
    cls = nmo.regularizer.ElasticNet

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_init_solver_name(self, solver_name):
        """Test ElasticNet acceptable solvers."""
        acceptable_solvers = [
            "ProximalGradient",
            "ProxSVRG",
        ]
        raise_exception = solver_name not in acceptable_solvers
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                nmo.glm.GLM(
                    regularizer=self.cls(),
                    solver_name=solver_name,
                    regularizer_strength=(1, 0.5),
                )
        else:
            nmo.glm.GLM(
                regularizer=self.cls(),
                solver_name=solver_name,
                regularizer_strength=(1, 0.5),
            )

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_set_solver_name_allowed(self, solver_name):
        """Test ElasticNet acceptable solvers."""
        acceptable_solvers = [
            "ProximalGradient",
            "ProxSVRG",
        ]
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=(1, 0.5))
        raise_exception = solver_name not in acceptable_solvers
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                model.set_params(solver_name=solver_name)
        else:
            model.set_params(solver_name=solver_name)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    @pytest.mark.parametrize("solver_kwargs", [{"tol": 10**-10}, {"tols": 10**-10}])
    def test_init_solver_kwargs(self, solver_kwargs, solver_name):
        """Test ElasticNetSolver acceptable kwargs."""
        regularizer = self.cls()
        raise_exception = "tols" in list(solver_kwargs.keys())
        if raise_exception:
            with pytest.raises(
                NameError, match="kwargs {'tols'} in solver_kwargs not a kwarg"
            ):
                nmo.glm.GLM(
                    regularizer=regularizer,
                    solver_name=solver_name,
                    solver_kwargs=solver_kwargs,
                    regularizer_strength=(1.0, 0.5),
                )
        else:
            nmo.glm.GLM(
                regularizer=regularizer,
                solver_name=solver_name,
                solver_kwargs=solver_kwargs,
                regularizer_strength=(1.0, 0.5),
            )

    def test_regularizer_strength_none(self):
        """Assert regularizer strength handled appropriately."""
        # if no strength given, set to (1.0, 0.5)
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer)

        assert model.regularizer_strength == (1.0, 0.5)

        # if changed to regularized, should go to None
        model.set_params(regularizer="UnRegularized", regularizer_strength=None)
        assert model.regularizer_strength is None

        # if changed back, set to (1.0, 0.5)
        model.regularizer = regularizer

        assert model.regularizer_strength == (1.0, 0.5)

    def test_regularizer_strength_float(self):
        """Assert regularizer ratio handled appropriately when only strength provided."""
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=0.6)
        assert model.regularizer_strength == (0.6, 0.5)

    @pytest.mark.parametrize(
        "regularizer_strength, expectation",
        [
            ((1.0, 0.5), does_not_raise()),
            ((1.0, 1.0), does_not_raise()),
            (
                (1.0, 0.0),
                pytest.raises(
                    ValueError,
                    match="Regularization ratio of 0 is not supported. Use Ridge regularization instead.",
                ),
            ),
            (
                (1.0, 1.1),
                pytest.raises(
                    ValueError,
                    match="Regularization ratio must be a number between 0 and 1.",
                ),
            ),
            (
                (1.0, -0.1),
                pytest.raises(
                    ValueError,
                    match="Regularization ratio must be a number between 0 and 1.",
                ),
            ),
            (
                (1.0, "bah"),
                pytest.raises(
                    ValueError,
                    match="Could not convert the regularizer strength and regularizer ratio",
                ),
            ),
            (
                (1.0, 0.5, 0.1),
                pytest.raises(
                    ValueError,
                    match="regularizer_strength must be a tuple of two floats",
                ),
            ),
        ],
    )
    def test_regularizer_ratio_setter(self, regularizer_strength, expectation):
        """Test that the regularizer ratio setter works as expected."""
        regularizer = self.cls()
        with expectation:
            nmo.glm.GLM(
                regularizer=regularizer, regularizer_strength=regularizer_strength
            )

    def test_get_params(self):
        """Test get_params() returns expected values."""
        regularizer = self.cls()

        assert regularizer.get_params() == {}

    @pytest.mark.parametrize("loss", [lambda a, b, c: 0, 1, None, {}])
    def test_loss_callable(self, loss):
        """Test that the loss function is a callable"""
        raise_exception = not callable(loss)
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=(1, 0.5))
        model._predict_and_compute_loss = loss
        if raise_exception:
            with pytest.raises(TypeError, match="The `loss` must be a Callable"):
                nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")
        else:
            nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_run_solver(self, solver_name, poissonGLM_model_instantiation):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        model.set_params(regularizer=self.cls(), regularizer_strength=(1, 0.5))
        model.solver_name = solver_name
        runner = model.instantiate_solver().solver_run
        runner((true_params[0] * 0.0, true_params[1]), X, y)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_run_solver_tree(self, solver_name, poissonGLM_model_instantiation_pytree):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation_pytree

        # set regularizer and solver name
        model.set_params(regularizer=self.cls(), regularizer_strength=(1, 0.5))
        model.solver_name = solver_name
        runner = model.instantiate_solver().solver_run
        runner(
            (jax.tree_util.tree_map(jnp.zeros_like, true_params[0]), true_params[1]),
            X.data,
            y,
        )

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    @pytest.mark.parametrize("reg_strength", [1.0, 0.5, 0.1])
    @pytest.mark.parametrize("reg_ratio", [1.0, 0.5, 0.2])
    def test_solver_match_statsmodels(
        self, solver_name, reg_strength, reg_ratio, poissonGLM_model_instantiation
    ):
        """Test that different solvers converge to the same solution."""
        jax.config.update("jax_enable_x64", True)
        # with jax.disable_jit():
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        # set precision to float64 for accurate matching of the results
        model.data_type = jnp.float64
        model.set_params(
            regularizer=self.cls(), regularizer_strength=(reg_strength, reg_ratio)
        )
        model.solver_name = solver_name
        model.solver_kwargs = {"tol": 10**-12, "maxiter": 10000}

        runner = model.instantiate_solver().solver_run
        weights, intercepts = runner((true_params[0] * 0.0, true_params[1]), X, y)[0]

        model.fit(X, y)
        # instantiate the glm with statsmodels
        glm_sm = sm.GLM(endog=y, exog=sm.add_constant(X), family=sm.families.Poisson())

        # regularize everything except intercept
        alpha_sm = np.ones(X.shape[1] + 1) * reg_strength
        alpha_sm[0] = 0

        # pure lasso = elastic net with L1 weight = 1
        res_sm = glm_sm.fit_regularized(
            method="elastic_net",
            alpha=alpha_sm,
            L1_wt=reg_ratio,
            cnvrg_tol=10**-12,
            zero_tol=1e-1000,
            maxiter=10000,
        )
        # compare params
        sm_params = res_sm.params
        glm_params = jnp.hstack((intercepts, weights.flatten()))
        assert np.allclose(sm_params, glm_params)

    def test_loss_convergence(self):
        """Test that penalized loss converges to the same value as statsmodels and the proximal operator."""
        jax.config.update("jax_enable_x64", True)
        # generate toy data
        np.random.seed(123)
        num_samples, num_features = 1000, 5
        X = np.random.normal(size=(num_samples, num_features))  # design matrix
        w = list(np.random.normal(size=(num_features,)))  # define some weights
        y = np.random.poisson(np.exp(X.dot(w)))  # observed counts

        # instantiate and fit GLM with ProximalGradient
        model_PG = nmo.glm.GLM(
            regularizer="ElasticNet",
            regularizer_strength=(1.0, 0.5),
            solver_name="ProximalGradient",
            solver_kwargs=dict(tol=10**-12, maxiter=10000),
        )
        model_PG.fit(X, y)
        glm_res = np.hstack((model_PG.intercept_, model_PG.coef_))

        # use the penalized loss function to solve optimization via Nelder-Mead
        penalized_loss = lambda p, x, y: model_PG.regularizer.penalized_loss(
            model_PG._predict_and_compute_loss, model_PG.regularizer_strength
        )(
            (
                p[1:],
                p[0].reshape(
                    1,
                ),
            ),
            x,
            y,
        )
        res = minimize(
            penalized_loss,
            [0] + w,
            args=(X, y),
            method="Nelder-Mead",
            tol=10**-12,
            options={"maxiter": 10000},
        )
        # regularize everything except intercept
        alpha_sm = np.ones(X.shape[1] + 1) * model_PG.regularizer_strength[0]
        alpha_sm[0] = 0

        # elastic net with
        glm_sm = sm.GLM(endog=y, exog=sm.add_constant(X), family=sm.families.Poisson())
        res_sm = glm_sm.fit_regularized(
            method="elastic_net",
            alpha=alpha_sm,
            L1_wt=0.5,
            cnvrg_tol=10**-12,
            zero_tol=1e-1000,
            maxiter=10000,
        )
        # assert weights are the same
        assert np.allclose(res.x, glm_res)
        assert np.allclose(res.x, res_sm.params)
        assert np.allclose(glm_res, res_sm.params)

    def test_elasticnet_pytree(self, poissonGLM_model_instantiation_pytree):
        """Check pytree X can be fit."""
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation_pytree
        model.set_params(
            regularizer=nmo.regularizer.ElasticNet(), regularizer_strength=1.0
        )
        model.solver_name = "ProximalGradient"
        model.fit(X, y)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    @pytest.mark.parametrize("reg_str", [0.001, 0.01, 0.1, 1, 10])
    def test_elasticnet_pytree_match(
        self,
        reg_str,
        solver_name,
        poissonGLM_model_instantiation_pytree,
        poissonGLM_model_instantiation,
    ):
        """Check pytree and array find same solution."""
        jax.config.update("jax_enable_x64", True)
        X, _, model, _, _ = poissonGLM_model_instantiation_pytree
        X_array, y, model_array, _, _ = poissonGLM_model_instantiation

        model.set_params(
            regularizer=nmo.regularizer.ElasticNet(), regularizer_strength=reg_str
        )
        model_array.set_params(
            regularizer=nmo.regularizer.ElasticNet(), regularizer_strength=reg_str
        )
        model.solver_name = solver_name
        model_array.solver_name = solver_name
        model.fit(X, y)
        model_array.fit(X_array, y)
        assert np.allclose(
            np.hstack(jax.tree_util.tree_leaves(model.coef_)), model_array.coef_
        )

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_solver_combination(self, solver_name, poissonGLM_model_instantiation):
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        model.set_params(regularizer=self.cls(), regularizer_strength=1.0)
        model.solver_name = solver_name
        model.fit(X, y)


class TestGroupLasso:
    cls = nmo.regularizer.GroupLasso

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_init_solver_name(self, solver_name):
        """Test GroupLasso acceptable solvers."""
        acceptable_solvers = [
            "ProximalGradient",
            "ProxSVRG",
        ]
        raise_exception = solver_name not in acceptable_solvers

        # create a valid mask
        mask = np.zeros((2, 10))
        mask[0, :5] = 1
        mask[1, 5:] = 1
        mask = jnp.asarray(mask)

        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                nmo.glm.GLM(
                    regularizer=self.cls(mask=mask),
                    solver_name=solver_name,
                    regularizer_strength=1,
                )
        else:
            nmo.glm.GLM(
                regularizer=self.cls(mask=mask),
                solver_name=solver_name,
                regularizer_strength=1,
            )

    @pytest.mark.parametrize(
        "solver_name",
        [
            "GradientDescent",
            "BFGS",
            "ProximalGradient",
            "AGradientDescent",
            1,
            "SVRG",
            "ProxSVRG",
        ],
    )
    def test_set_solver_name_allowed(self, solver_name):
        """Test GroupLassoSolver acceptable solvers."""
        acceptable_solvers = [
            "ProximalGradient",
            "ProxSVRG",
        ]
        # create a valid mask
        mask = np.zeros((2, 10))
        mask[0, :5] = 1
        mask[1, 5:] = 1
        mask = jnp.asarray(mask)
        regularizer = self.cls(mask=mask)
        raise_exception = solver_name not in acceptable_solvers
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=1)
        if raise_exception:
            with pytest.raises(
                ValueError, match=f"The solver: {solver_name} is not allowed for "
            ):
                model.set_params(solver_name=solver_name)
        else:
            model.set_params(solver_name=solver_name)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    @pytest.mark.parametrize("solver_kwargs", [{"tol": 10**-10}, {"tols": 10**-10}])
    def test_init_solver_kwargs(self, solver_name, solver_kwargs):
        """Test GroupLasso acceptable kwargs."""
        raise_exception = "tols" in list(solver_kwargs.keys())

        # create a valid mask
        mask = np.zeros((2, 10))
        mask[0, :5] = 1
        mask[0, 1:] = 1
        mask = jnp.asarray(mask)

        regularizer = self.cls(mask=mask)

        if raise_exception:
            with pytest.raises(
                NameError, match="kwargs {'tols'} in solver_kwargs not a kwarg"
            ):
                nmo.glm.GLM(
                    regularizer=regularizer,
                    solver_name=solver_name,
                    solver_kwargs=solver_kwargs,
                    regularizer_strength=1.0,
                )
        else:
            nmo.glm.GLM(
                regularizer=regularizer,
                solver_name=solver_name,
                solver_kwargs=solver_kwargs,
                regularizer_strength=1.0,
            )

    def test_regularizer_strength_none(self):
        """Assert regularizer strength handled appropriately."""
        # if no strength given, should set to 1.0
        regularizer = self.cls()
        model = nmo.glm.GLM(regularizer=regularizer)

        assert model.regularizer_strength == 1.0

        # if changed to regularized, should go to None
        model.set_params(regularizer="UnRegularized", regularizer_strength=None)
        assert model.regularizer_strength is None

        # if changed back, should set to 1.0
        model.regularizer = regularizer

        assert model.regularizer_strength == 1.0

    def test_get_params(self):
        """Test get_params() returns expected values."""
        regularizer = self.cls()

        assert regularizer.get_params() == {"mask": None}

    @pytest.mark.parametrize("loss", [lambda a, b, c: 0, 1, None, {}])
    def test_loss_callable(self, loss):
        """Test that the loss function is a callable"""
        raise_exception = not callable(loss)

        # create a valid mask
        mask = np.zeros((2, 10))
        mask[0, :5] = 1
        mask[1, 5:] = 1
        mask = jnp.asarray(mask)

        regularizer = self.cls(mask=mask)
        model = nmo.glm.GLM(regularizer=regularizer, regularizer_strength=1.0)
        model._predict_and_compute_loss = loss

        if raise_exception:
            with pytest.raises(TypeError, match="The `loss` must be a Callable"):
                nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")
        else:
            nmo.utils.assert_is_callable(model._predict_and_compute_loss, "loss")

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_run_solver(self, solver_name, poissonGLM_model_instantiation):
        """Test that the solver runs."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # create a valid mask
        mask = np.zeros((2, X.shape[1]))
        mask[0, :2] = 1
        mask[1, 2:] = 1
        mask = jnp.asarray(mask)

        model.set_params(regularizer=self.cls(mask=mask), regularizer_strength=1.0)
        model.solver_name = solver_name

        model.instantiate_solver()
        model.solver_run((true_params[0] * 0.0, true_params[1]), X, y)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_init_solver(self, solver_name, poissonGLM_model_instantiation):
        """Test that the solver initialization returns a state."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # create a valid mask
        mask = np.zeros((2, X.shape[1]))
        mask[0, :2] = 1
        mask[1, 2:] = 1
        mask = jnp.asarray(mask)

        model.set_params(regularizer=self.cls(mask=mask), regularizer_strength=1.0)
        model.solver_name = solver_name

        model.instantiate_solver()
        state = model.solver_init_state(true_params, X, y)
        # asses that state is a NamedTuple by checking tuple type and the availability of some NamedTuple
        # specific namespace attributes
        assert isinstance(state, tuple | eqx.Module)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_update_solver(self, solver_name, poissonGLM_model_instantiation):
        """Test that the solver initialization returns a state."""

        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # create a valid mask
        mask = np.zeros((2, X.shape[1]))
        mask[0, :2] = 1
        mask[1, 2:] = 1
        mask = jnp.asarray(mask)

        model.set_params(regularizer=self.cls(mask=mask), regularizer_strength=1.0)
        model.solver_name = solver_name

        model.instantiate_solver()

        state = model.solver_init_state((true_params[0] * 0.0, true_params[1]), X, y)

        # ProxSVRG needs the full gradient at the anchor point to be initialized
        # so here just set it to xs, which is not correct, but fine shape-wise
        if solver_name == "ProxSVRG":
            state = state._replace(full_grad_at_reference_point=state.reference_point)

        params, state = model.solver_update(true_params, state, X, y)
        # asses that state is a NamedTuple by checking tuple type and the availability of some NamedTuple
        # specific namespace attributes
        assert isinstance(state, tuple | eqx.Module)

        # check params struct and shapes
        assert jax.tree_util.tree_structure(params) == jax.tree_util.tree_structure(
            true_params
        )
        assert all(
            jax.tree_util.tree_leaves(params)[k].shape == p.shape
            for k, p in enumerate(jax.tree_util.tree_leaves(true_params))
        )

    @pytest.mark.parametrize("n_groups_assign", [0, 1, 2])
    def test_mask_validity_groups(
        self, n_groups_assign, poissonGLM_model_instantiation_group_sparse
    ):
        """Test that mask assigns at most 1 group to each weight."""
        raise_exception = n_groups_assign > 1
        (
            X,
            y,
            model,
            true_params,
            firing_rate,
            _,
        ) = poissonGLM_model_instantiation_group_sparse

        # create a valid mask
        mask = np.zeros((2, X.shape[1]))
        mask[0, :2] = 1
        mask[1, 2:] = 1

        # change assignment
        if n_groups_assign == 0:
            mask[:, 3] = 0
        elif n_groups_assign == 2:
            mask[:, 3] = 1

        mask = jnp.asarray(mask)

        if raise_exception:
            with pytest.raises(
                ValueError, match="Incorrect group assignment. Some of the features"
            ):
                model.set_params(
                    regularizer=self.cls(mask=mask), regularizer_strength=1.0
                )
        else:
            model.set_params(regularizer=self.cls(mask=mask), regularizer_strength=1.0)

    @pytest.mark.parametrize("set_entry", [0, 1, -1, 2, 2.5])
    def test_mask_validity_entries(self, set_entry, poissonGLM_model_instantiation):
        """Test that mask is composed of 0s and 1s."""
        raise_exception = set_entry not in {0, 1}
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # create a valid mask
        mask = np.zeros((2, X.shape[1]))
        mask[0, :2] = 1
        mask[1, 2:] = 1
        # assign an entry
        mask[1, 2] = set_entry
        mask = jnp.asarray(mask, dtype=jnp.float32)

        if raise_exception:
            with pytest.raises(ValueError, match="Mask elements be 0s and 1s"):
                model.set_params(
                    regularizer=self.cls(mask=mask), regularizer_strength=1.0
                )
        else:
            model.set_params(regularizer=self.cls(mask=mask), regularizer_strength=1.0)

    @pytest.mark.parametrize("n_dim", [0, 1, 2, 3])
    def test_mask_dimension_1(self, n_dim, poissonGLM_model_instantiation):
        """Test that mask is composed of 0s and 1s."""

        raise_exception = n_dim != 2
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # create a valid mask
        if n_dim == 0:
            mask = np.array([])
        elif n_dim == 1:
            mask = np.ones((1,))
        elif n_dim == 2:
            mask = np.zeros((2, X.shape[1]))
            mask[0, :2] = 1
            mask[1, 2:] = 1
        else:
            mask = np.zeros((2, X.shape[1]) + (1,) * (n_dim - 2))
            mask[0, :2] = 1
            mask[1, 2:] = 1

        mask = jnp.asarray(mask, dtype=jnp.float32)

        if raise_exception:
            with pytest.raises(ValueError, match="`mask` must be 2-dimensional"):
                model.set_params(
                    regularizer=self.cls(mask=mask), regularizer_strength=1.0
                )
        else:
            model.set_params(regularizer=self.cls(mask=mask), regularizer_strength=1.0)

    @pytest.mark.parametrize("n_groups", [0, 1, 2])
    def test_mask_n_groups(self, n_groups, poissonGLM_model_instantiation):
        """Test that mask has at least 1 group."""
        raise_exception = n_groups < 1
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # create a mask
        mask = np.zeros((n_groups, X.shape[1]))
        if n_groups > 0:
            for i in range(n_groups - 1):
                mask[i, i : i + 1] = 1
            mask[-1, n_groups - 1 :] = 1

        mask = jnp.asarray(mask, dtype=jnp.float32)

        if raise_exception:
            with pytest.raises(ValueError, match=r"Empty mask provided! Mask has "):
                model.set_params(
                    regularizer=self.cls(mask=mask), regularizer_strength=1.0
                )
        else:
            model.set_params(regularizer=self.cls(mask=mask), regularizer_strength=1.0)

    def test_group_sparsity_enforcement(
        self, poissonGLM_model_instantiation_group_sparse
    ):
        """Test that group lasso works on a simple dataset."""
        (
            X,
            y,
            model,
            true_params,
            firing_rate,
            _,
        ) = poissonGLM_model_instantiation_group_sparse
        zeros_true = true_params[0].flatten() == 0
        mask = np.zeros((2, X.shape[1]))
        mask[0, zeros_true] = 1
        mask[1, ~zeros_true] = 1
        mask = jnp.asarray(mask, dtype=jnp.float32)

        model.set_params(regularizer=self.cls(mask=mask), regularizer_strength=1.0)
        model.solver_name = "ProximalGradient"

        runner = model.instantiate_solver().solver_run
        params, _ = runner((true_params[0] * 0.0, true_params[1]), X, y)

        zeros_est = params[0] == 0
        if not np.all(zeros_est == zeros_true):
            raise ValueError("GroupLasso failed to zero-out the parameter group!")

    ###########
    # Test mask from set_params
    ###########
    @pytest.mark.parametrize("n_groups_assign", [0, 1, 2])
    def test_mask_validity_groups_set_params(
        self, n_groups_assign, poissonGLM_model_instantiation_group_sparse
    ):
        """Test that mask assigns at most 1 group to each weight."""
        raise_exception = n_groups_assign > 1
        (
            X,
            y,
            model,
            true_params,
            firing_rate,
            _,
        ) = poissonGLM_model_instantiation_group_sparse

        # create a valid mask
        mask = np.zeros((2, X.shape[1]))
        mask[0, :2] = 1
        mask[1, 2:] = 1
        regularizer = self.cls(mask=mask)

        # change assignment
        if n_groups_assign == 0:
            mask[:, 3] = 0
        elif n_groups_assign == 2:
            mask[:, 3] = 1

        mask = jnp.asarray(mask)

        if raise_exception:
            with pytest.raises(
                ValueError, match="Incorrect group assignment. Some of the features"
            ):
                regularizer.set_params(mask=mask)
        else:
            regularizer.set_params(mask=mask)

    @pytest.mark.parametrize("set_entry", [0, 1, -1, 2, 2.5])
    def test_mask_validity_entries_set_params(
        self, set_entry, poissonGLM_model_instantiation
    ):
        """Test that mask is composed of 0s and 1s."""
        raise_exception = set_entry not in {0, 1}
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # create a valid mask
        mask = np.zeros((2, X.shape[1]))
        mask[0, :2] = 1
        mask[1, 2:] = 1
        regularizer = self.cls(mask=mask)

        # assign an entry
        mask[1, 2] = set_entry
        mask = jnp.asarray(mask, dtype=jnp.float32)

        if raise_exception:
            with pytest.raises(ValueError, match="Mask elements be 0s and 1s"):
                regularizer.set_params(mask=mask)
        else:
            regularizer.set_params(mask=mask)

    @pytest.mark.parametrize("n_dim", [0, 1, 2, 3])
    def test_mask_dimension(self, n_dim, poissonGLM_model_instantiation):
        """Test that mask is composed of 0s and 1s."""

        raise_exception = n_dim != 2
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        valid_mask = np.zeros((2, X.shape[1]))
        valid_mask[0, :1] = 1
        valid_mask[1, 1:] = 1
        regularizer = self.cls(mask=valid_mask)

        # create a mask
        if n_dim == 0:
            mask = np.array([])
        elif n_dim == 1:
            mask = np.ones((1,))
        elif n_dim == 2:
            mask = np.zeros((2, X.shape[1]))
            mask[0, :2] = 1
            mask[1, 2:] = 1
        else:
            mask = np.zeros((2, X.shape[1]) + (1,) * (n_dim - 2))
            mask[0, :2] = 1
            mask[1, 2:] = 1

        mask = jnp.asarray(mask, dtype=jnp.float32)

        if raise_exception:
            with pytest.raises(ValueError, match="`mask` must be 2-dimensional"):
                regularizer.set_params(mask=mask)
        else:
            regularizer.set_params(mask=mask)

    @pytest.mark.parametrize("n_groups", [0, 1, 2])
    def test_mask_n_groups_set_params(self, n_groups, poissonGLM_model_instantiation):
        """Test that mask has at least 1 group."""
        raise_exception = n_groups < 1
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        valid_mask = np.zeros((2, X.shape[1]))
        valid_mask[0, :1] = 1
        valid_mask[1, 1:] = 1
        regularizer = self.cls(mask=valid_mask)

        # create a mask
        mask = np.zeros((n_groups, X.shape[1]))
        if n_groups > 0:
            for i in range(n_groups - 1):
                mask[i, i : i + 1] = 1
            mask[-1, n_groups - 1 :] = 1

        mask = jnp.asarray(mask, dtype=jnp.float32)

        if raise_exception:
            with pytest.raises(ValueError, match=r"Empty mask provided! Mask has "):
                regularizer.set_params(mask=mask)
        else:
            regularizer.set_params(mask=mask)

    def test_mask_none(self, poissonGLM_model_instantiation):
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation

        # with pytest.warns(UserWarning):
        model.regularizer = self.cls(mask=np.ones((1, X.shape[1])).astype(float))
        model.solver_name = "ProximalGradient"
        model.fit(X, y)

    @pytest.mark.parametrize("solver_name", ["ProximalGradient", "ProxSVRG"])
    def test_solver_combination(self, solver_name, poissonGLM_model_instantiation):
        X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
        model.set_params(
            regularizer=self.cls(mask=np.ones((1, X.shape[1])).astype(float)),
            regularizer_strength=(
                None if self.cls == nmo.regularizer.UnRegularized else 1.0
            ),
        )
        model.solver_name = solver_name
        model.fit(X, y)


def test_available_regularizer_match():
    """Test matching of the two regularizer lists."""
    assert set(nmo._regularizer_builder.AVAILABLE_REGULARIZERS) == set(
        nmo.regularizer.__dir__()
    )



"""
Utility function for composite basis.

Collection of functions that transverse the composite basis tree
with no to minimal re
"""

import re
from copy import deepcopy
from functools import wraps
from typing import TYPE_CHECKING, Any, List, Tuple

import numpy as np
from numpy.typing import NDArray

from .._inspect_utils.inspect_utils import count_positional_and_var_args

if TYPE_CHECKING:
    from ._basis_mixin import AtomicBasisMixin, BasisMixin
    from ._custom_basis import CustomBasis

__PUBLIC_BASES__ = [
    "IdentityEval",
    "HistoryConv",
    "MSplineEval",
    "MSplineConv",
    "BSplineEval",
    "BSplineConv",
    "CyclicBSplineEval",
    "CyclicBSplineConv",
    "FourierEval",
    "RaisedCosineLinearEval",
    "RaisedCosineLinearConv",
    "RaisedCosineLogEval",
    "RaisedCosineLogConv",
    "OrthExponentialEval",
    "OrthExponentialConv",
    "AdditiveBasis",
    "MultiplicativeBasis",
    "CustomBasis",
]


def add_docstring(method_name, cls):
    """Prepend super-class docstrings."""
    attr = getattr(cls, method_name, None)
    if attr is None:
        raise AttributeError(f"{cls.__name__} has no attribute {method_name}!")
    doc = attr.__doc__

    # Decorator to add the docstring
    def wrapper(func):
        func.__doc__ = "\n".join([doc, func.__doc__])  # Combine docstrings
        return func

    return wrapper


def _iterate_over_components(basis: "BasisMixin"):
    components = (
        basis._iterate_over_components()
        if hasattr(basis, "_iterate_over_components")
        else [basis]
    )
    yield from components


def _get_root(bas: "BasisMixin"):
    """Get the basis root."""

    parent = bas
    while hasattr(parent, "_parent") and parent._parent is not None:
        parent = parent._parent
    return parent


def _has_default_label(bas: "BasisMixin"):
    """Check for default label.

    The check either use the property (if it is a nemos basis), or the class name
    (if it is a custom user defined basis).
    """
    if hasattr(bas, "_has_default_label"):
        return bas._has_default_label
    else:
        label = getattr(bas, "label", bas.__class__.__name__)
        return re.match(rf"^{bas.__class__.__name__}(_\d+)?$", label)


def _recompute_class_default_labels(
    bas: "BasisMixin",
):
    """
    Recompute all labels matching default for self.

    Parameters
    ----------
    bas:
        Basis component calling the method
    """
    class_name = bas.__class__.__name__
    pattern = re.compile(rf"^{class_name}(_\d+)?$")
    root = _get_root(bas)
    bas_id = 0
    # if root is one of our bases it will have the iteration method, if custom from user
    # I assume it is atomic
    for comp_bas in _iterate_over_components(root):
        if re.match(pattern, comp_bas._label):
            comp_bas._label = f"{class_name}_{bas_id}" if bas_id else class_name
            bas_id += 1


def _recompute_all_default_labels(root: "BasisMixin") -> "BasisMixin":
    """Recompute default all labels."""
    updated = []
    for bas in _iterate_over_components(root):
        if _has_default_label(bas) and bas.__class__.__name__ not in updated:
            _recompute_class_default_labels(bas)
            updated.append(bas.__class__.__name__)
    return root


def _update_label_from_root(bas: "BasisMixin", class_name: str, class_label: str):
    """
    Subtract 1 to each matching default label with higher ID then current.

    Parameters
    ----------
    class_name : str
        Class name of the component that is setting a new label.
    class_label : str
        Current component label.
    """
    pattern = re.compile(rf"^{class_name}(_\d+)?$")
    match = re.match(pattern, class_label)
    if match is None:
        return
    # get the "ID" of the label
    current_id = int(match.group(1)[1:]) if match.group(1) else 0
    # subtract one to the ID of any other default label with ID > current_id
    root = _get_root(bas)
    for bas in _iterate_over_components(root):
        match = re.match(pattern, bas._label)
        if match:
            bas_id = int(match.group(1)[1:]) if match.group(1) else 0
            bas_id = bas_id - 1 if bas_id > current_id else bas_id
            bas._label = f"{class_name}_{bas_id}" if bas_id else class_name


def _composite_basis_setter_logic(new: "BasisMixin", current: "BasisMixin"):
    """Setter logic for composite basis."""
    # Carry-on label if possible
    if _has_default_label(new) and not _has_default_label(current):
        try:
            new.label = getattr(current, "label", current.__class__.__name__)
        except ValueError:
            pass  # If label is in use, ignore

    # add a parent to the new basis
    new._parent = getattr(current, "_parent", None)

    # Carry-on input shape info if dimensions match
    for attr in ("_input_shape_product", "_input_shape_"):
        if getattr(new, attr, None) is None and getattr(
            new, "_n_input_dimensionality", None
        ) == getattr(current, "_n_input_dimensionality", None):
            setattr(new, attr, getattr(current, attr, None))
    return new


def _atomic_basis_label_setter_logic(
    bas: "AtomicBasisMixin | CustomBasis", new_label: str
) -> Exception | None:
    """Setter logic for atomic basis."""
    # check default cases
    current_label = getattr(bas, "_label", None)
    if new_label == current_label:
        return

    elif new_label is None:
        # check if already default
        bas._label = bas.__class__.__name__
        _recompute_class_default_labels(bas)
        return

        # raise error in case label is not string.
    elif not isinstance(new_label, str):
        return TypeError(
            f"'label' must be a string. Type {type(new_label)} was provided instead."
        )

    else:
        # check if label matches class-name plus identifier
        match = re.match(r"(.+)?_\d+$", new_label)
        check_string = match.group(1) if match else None
        check_string = check_string if check_string in __PUBLIC_BASES__ else new_label
        if check_string == bas.__class__.__name__:
            bas._label = check_string
            _recompute_class_default_labels(bas)
            return

    root = _get_root(bas)
    current_labels = (
        root._generate_subtree_labels() if root is not bas else [current_label]
    )
    if (check_string in current_labels) or (check_string in __PUBLIC_BASES__):
        if check_string in __PUBLIC_BASES__:
            msg = f"Cannot assign '{new_label}' to a basis of class {bas.__class__.__name__}."
        else:
            msg = f"Label '{new_label}' is already in use. When user-provided, label must be unique."
        return ValueError(msg)
    else:
        # check if current is the default label
        # if that's true, since the new label is not a default label
        # update all other default label names.
        _update_label_from_root(
            bas, bas.__class__.__name__, getattr(bas, "_label", bas.label)
        )
        bas._label = new_label
    return


def infer_input_dimensionality(bas: "BasisMixin") -> int:
    """Infer input dimensionality from compute_features signature.

    If `_n_input_dimensionality` return the attribute, otherwise return
    the number of fixed arguments in `compute_features`.
    """
    n_input_dim = getattr(bas, "_n_input_dimensionality", None)
    if n_input_dim is None:
        # infer from compute_features (facilitate custom basis compatibility).
        # assume compute_features is always implemented.
        if hasattr(bas, "funcs"):
            funcs = bas.funcs
            dims = [count_positional_and_var_args(f)[0] for f in funcs]
            if len(set(dims)) != 1:
                raise ValueError(
                    "Each function provided to ``funcs`` in ``CustomBasis`` must accept the same "
                    "number of input time series."
                )
            n_input_dim = dims[0]
        else:
            funcs = [bas.compute_features] if hasattr(bas, "compute_features") else []
            n_input_dim = sum(count_positional_and_var_args(f)[0] for f in funcs)
    return n_input_dim


def generate_basis_label_pair(bas: "BasisMixin"):
    """Generate all labels and basis in a basis tree."""
    if hasattr(bas, "basis1") and hasattr(bas, "basis2"):
        for label, sub_bas in generate_basis_label_pair(bas.basis1):
            yield label, sub_bas
        for label, sub_bas in generate_basis_label_pair(bas.basis2):
            yield label, sub_bas
    yield getattr(bas, "label", bas.__class__.__name__), bas


def generate_composite_basis_labels(bas: "BasisMixin", type_label: str) -> str:
    """Generate all labels in a basis tree."""
    if hasattr(bas, "basis1") and hasattr(bas, "basis2"):
        if type_label == "all" or bas._label:
            yield bas.label

        # generator for sub-basis
        def generate_labels(sub_basis):
            generate_subtree = getattr(sub_basis, "_generate_subtree_labels", None)
            if generate_subtree:
                yield from generate_subtree(type_label)
            elif type_label == "all" or not _has_default_label(sub_basis):
                yield sub_basis.label

        yield from generate_labels(bas.basis1)
        yield from generate_labels(bas.basis2)

    else:
        if type_label == "all" or (not _has_default_label(bas)):
            yield getattr(bas, "label", bas.__class__.__name__)


def label_setter(bas: "BasisMixin", label: str | None) -> None | ValueError:
    """Label setter logic for any basis."""
    if not hasattr(bas, "basis1") or not hasattr(bas, "basis2"):
        return _atomic_basis_label_setter_logic(bas, label)

    # composite basis setter logic.
    reset = (
        (label is None)
        or (label == bas._generate_label())
        or (label == bas.__class__.__name__)
    )
    error = None
    if reset:
        label = None
    else:
        label = str(label)
        if label in _get_root(bas)._generate_subtree_labels():
            error = ValueError(
                f"Label '{label}' is already in use. When user-provided, label must be unique."
            )
        elif label in __PUBLIC_BASES__ and label != bas.__class__.__name__:
            error = ValueError(
                f"Cannot set basis label '{label}' for basis of type {type(bas)}."
            )
    if not error:
        bas._label = label
    return error


def _check_valid_shape_tuple(*shapes):
    for shape in shapes:
        if shape is not None and not all(isinstance(i, int) for i in shape):
            raise ValueError(
                f"The tuple provided contains non integer values. Tuple: {shape}."
            )


def transform_to_shape(xi):
    if isinstance(xi, tuple):
        shape = xi
    elif isinstance(xi, int):
        shape = () if xi == 1 else (xi,)
    else:
        shape = xi if xi is None else xi.shape[1:]
    return shape


def set_input_shape_atomic(
    bas: "AtomicBasisMixin | CustomBasis", *xis: int | tuple[int, ...] | NDArray | None
) -> "AtomicBasisMixin":
    """Set input shape attributes for atomic basis."""
    # reset all to none
    if all(xi is None for xi in xis):
        bas._input_shape_ = None
        bas._input_shape_product = None
        return bas
    # otherwise apply shape
    shapes = []
    n_inputs = ()
    for xi in xis:
        n_inputs = (*n_inputs, int(np.prod(xi)))
        shapes.append(xi)

    bas._input_shape_ = shapes

    # total number of input time series. Used  for slicing and reshaping
    bas._input_shape_product = n_inputs
    return bas


def _have_unique_shapes(inputs: List[NDArray] | List[Tuple]):
    if inputs is None:
        return True

    # Extract shapes whether inputs are arrays or shape tuples
    if hasattr(inputs[0], "shape"):
        shapes = {x.shape for x in inputs}
        context = "was evaluated with"
    else:
        shapes = set(inputs)
        context = "was configured with"
    return len(shapes) == 1, shapes, context


def _check_unique_shapes(inputs: List[NDArray] | List[Tuple], basis: "BasisMixin"):
    """Check that all inputs have the same shape.

    Parameters
    ----------
    inputs: Either a list of shape tuples or actual arrays
    basis: The basis object for error context

    Raises
    ------
    ValueError:
        If all the arrays have the same shapes or all the tuples are the same.
    """
    have_unique_shape, shapes, context = _have_unique_shapes(inputs)
    if not have_unique_shape:
        raise ValueError(
            f"{basis.__class__.__name__} requires all inputs to have the same shape. "
            f"The basis {context} input shapes: {shapes}.\n{basis}"
        )


def set_input_shape(bas, *xi, allow_inputs_of_different_shape=True):
    """Set input shape.

    Set input shape logic, compatible with all bases (composite, atomic, and custom).
    """

    # use 1 as default or number of non-variable args
    n_args = (
        count_positional_and_var_args(bas.compute_features)[0]
        if hasattr(bas, "compute_features")
        else 1
    )
    # get the attribute if available
    n_input_dim = getattr(bas, "_n_input_dimensionality", n_args)

    if len(xi) == 1 and xi[0] is None:
        xi = (None,) * n_input_dim

    elif len(xi) != n_input_dim:
        expected_inputs = getattr(bas, "_n_input_dimensionality", 1)
        raise ValueError(
            f"set_input_shape expects {expected_inputs} input"
            f"{'s' if expected_inputs > 1 else ''}, but {len(xi)} were provided."
        )

    original_xi = xi
    try:
        xi = [transform_to_shape(x) for x in xi]
    except Exception as e:
        raise ValueError(
            f"Cannot convert inputs ``{original_xi}`` to shape tuple."
        ) from e

    if not allow_inputs_of_different_shape:
        _check_unique_shapes(xi, bas)

    _check_valid_shape_tuple(*xi)

    if not hasattr(bas, "basis1"):
        return set_input_shape_atomic(bas, *xi)

    # here we can assume it is a composite basis
    # grab the input dimensionality
    n_args_1, _ = (
        count_positional_and_var_args(bas.basis1.compute_features)
        if hasattr(bas.basis1, "compute_features")
        else 1
    )
    n_input_dim_1 = getattr(bas.basis1, "_n_input_dimensionality", n_args_1)

    out1 = set_input_shape(bas.basis1, *xi[:n_input_dim_1])
    out2 = set_input_shape(bas.basis2, *xi[n_input_dim_1:])

    # out1 and out2 will have an _input_shape_product set by the "set_input_shape_atomic" method.
    # here is safe to use the attribute.
    if out1._input_shape_product is not None and out2._input_shape_product is not None:
        bas._input_shape_product = (
            *out1._input_shape_product,
            *out2._input_shape_product,
        )
    else:
        bas._input_shape_product = None
    return bas


def unpack_shapes(basis) -> Tuple:
    """Generate input shapes for any basis type.

    Generate input shape independently of the basis kind (i.e. works
    for composite, custom, and atomic bases).
    """
    if hasattr(basis, "_input_shape_") and hasattr(basis, "input_shape"):
        if basis._input_shape_ is None:
            yield None
        else:
            yield from basis._input_shape_
    elif hasattr(basis, "input_shape"):
        yield basis.input_shape
    else:
        yield from [None] * infer_input_dimensionality(basis)


def list_shapes(basis) -> List[Tuple | None]:
    """List input shape for all components."""
    if basis is None:
        return [None]
    components = getattr(basis, "_iterate_over_components", lambda: [basis])()
    return [shape for comp in components for shape in unpack_shapes(comp)]


def get_input_shape(bas: "BasisMixin") -> List[Tuple | None]:
    """Get the input shape of a composite basis.

    Get input shape from composition of basis, including user defined ones.
    The function treats any bases without a `_iterate_over_components` methods as an
    atomic basis.

    The input shape for user-defined bases is retrieved with _input_shape_ property that
    is set at runtime when `compute_features` or `set_input_shape` is called. If the
    number of inputs is not set yet, it returns a list of None.
    """
    input_dim = infer_input_dimensionality(bas)
    if input_dim == 1:
        ishape = getattr(bas, "_input_shape_", [None])
        return ishape if ishape is not None else [None]

    elif not hasattr(bas, "basis1") and hasattr(bas, "_input_shape_"):
        return [None] * input_dim if bas._input_shape_ is None else bas._input_shape_

    basis1 = getattr(bas, "_basis1", None)
    basis2 = getattr(bas, "_basis2", None)

    # Handle cases where one or both bases are missing
    if basis1 is None and basis2 is None:
        return [None, None]
    if basis1 is None:
        return [None] + list_shapes(basis2)
    if basis2 is None:
        return list_shapes(basis1) + [None]

    # If both bases exist, return combined shapes
    return list_shapes(basis1) + list_shapes(basis2)


def is_basis_like(putative_basis: Any, sklearn_compatibility=False) -> bool:
    """Check if putative_basis respect nemos basis API."""
    is_basis = hasattr(putative_basis, "compute_features")
    if sklearn_compatibility:
        is_basis &= hasattr(putative_basis, "get_params") and hasattr(
            putative_basis, "set_params"
        )
    return is_basis


def multiply_basis_by_integer(bas: "BasisMixin", mul: int) -> "BasisMixin":
    """Multiplication by integer logic for bases."""
    if mul <= 0:
        raise ValueError(
            "Basis multiplication error. Integer multiplicative factor must be positive, "
            f"{mul} provided instead."
        )
    elif not all(_has_default_label(b) for _, b in generate_basis_label_pair(bas)):
        raise ValueError(
            "Cannot multiply by an integer a basis including a user-defined labels "
            "(because then they won't be unique). Set labels after multiplication."
        )
    # default case
    if mul == 1:
        # __sklearn_clone__ reset the parent to None in case bas.basis1 * 1
        # (deepcopy would not)
        copy_ = getattr(bas.__class__, "__sklearn_clone__", deepcopy)
        bas = copy_(bas)

        # if deepcopy was called (custom basis used in composition)
        # reset _parent if it exists and is not None
        if hasattr(bas, "_parent") and bas._parent is not None:
            bas._parent = None
        _recompute_all_default_labels(bas)
        return bas

    # parent is set to None at init for add and updated for self.
    add = bas + bas
    with add._set_shallow_copy(True):
        for _ in range(2, mul):
            add += deepcopy(bas)
    _recompute_all_default_labels(add)
    return add


def raise_basis_to_power(bas: "BasisMixin", exponent: int) -> "BasisMixin":
    """Power of basis by integer."""
    if not isinstance(exponent, int):
        raise TypeError("Basis exponent should be an integer!")

    if exponent <= 0:
        raise ValueError("Basis exponent should be a non-negative integer!")
    elif not all(_has_default_label(b) for _, b in generate_basis_label_pair(bas)):
        raise ValueError(
            "Cannot calculate the power of a basis including a user-defined labels "
            "(because then they won't be unique). Set labels after exponentiation."
        )

    # default case
    if exponent == 1:
        # __sklearn_clone__ reset the parent to None in case bas.basis1 ** 1
        # (deepcopy would not)
        copy_ = getattr(bas.__class__, "__sklearn_clone__", deepcopy)
        bas = copy_(bas)

        # if deepcopy was called (custom basis used in composition)
        # reset _parent if it exists and it is not None
        if hasattr(bas, "_parent") and bas._parent is not None:
            bas._parent = None

        _recompute_all_default_labels(bas)
        return bas

    mul = bas * bas
    with mul._set_shallow_copy(True):
        for _ in range(2, exponent):
            mul *= deepcopy(bas)
    _recompute_all_default_labels(mul)
    return mul


def is_transformer(bas: Any):
    """Check if it conforms to transformer API."""
    return hasattr(bas, "basis") and hasattr(bas, "fit_transform")


def promote_to_transformer(method):
    """Apply operations to basis within transformer and transform output."""

    @wraps(method)
    def wrapper(*args, **kwargs):
        # if it looks like a transformer, pull out basis
        args_bas = (b.basis if is_transformer(b) else b for b in args)
        kwargs_bas = {k: b.basis if is_transformer(b) else b for k, b in kwargs.items()}
        out = method(*args_bas, **kwargs_bas)
        any_transformer = any(
            (
                *(is_transformer(a) for a in args),
                *(is_transformer(b) for b in kwargs.values()),
            )
        )
        if any_transformer and hasattr(out, "to_transformer"):
            return out.to_transformer()
        return out

    return wrapper



"""Utility functions for creating regularizer object."""

from .regularizer import ElasticNet, GroupLasso, Lasso, Ridge, UnRegularized

AVAILABLE_REGULARIZERS = ["UnRegularized", "Ridge", "Lasso", "GroupLasso", "ElasticNet"]

# Mapping for O(1) lookup
_REGULARIZER_MAP = {
    "UnRegularized": UnRegularized,
    "Ridge": Ridge,
    "Lasso": Lasso,
    "GroupLasso": GroupLasso,
    "ElasticNet": ElasticNet,
    None: UnRegularized,  # Handle None case
}


def instantiate_regularizer(name: str | None, **kwargs):
    """
    Create a regularizer from a given name.

    Parameters
    ----------
    name :
        The string name of the regularizer to create.
        Must be one of: 'UnRegularized', 'Ridge', 'Lasso','GroupLasso', None.
        If set to None, it will behave as 'Unregularized'.
    **kwargs :
        Additional keyword arguments are passed to the regularizer constructor.

    Returns
    -------
    :
        The regularizer instance with default parameters.

    Raises
    ------
    ValueError
        If the `name` provided does not match to any available regularizer.
    """
    # If the name contains a module path, extract the class name
    if name:
        name = name.split("nemos.regularizer.")[-1]

    if name in _REGULARIZER_MAP:
        return _REGULARIZER_MAP[name](**kwargs)
    else:
        raise ValueError(
            f"Unknown regularizer: {name}. "
            f"Regularizer must be one of {AVAILABLE_REGULARIZERS} or their full module path."
        )
